# -*- coding: utf-8 -*-
# pylint: disable=all
"""
Least Cost Xmission Command Line Interface

TODO - add cmd line doc

"""
import os
import sys
import time
import click
import logging
import warnings
import pandas as pd
import geopandas as gpd
from typing import List

from rex.utilities.loggers import init_mult, create_dirs, init_logger
from rex.utilities.cli_dtypes import STR, INTLIST, INT, FLOAT
from rex.utilities.hpc import SLURM
from rex.utilities.utilities import get_class_properties, dict_str_load

from reVX import __version__
from reVX.config.least_cost_xmission import LeastCostXmissionConfig
from reVX.least_cost_xmission.least_cost_xmission import (LeastCostXmission,
                                                          RegionalXmission)
from reVX.least_cost_xmission.config.constants import (CELL_SIZE,
                                                       TRANS_LINE_CAT,
                                                       LOAD_CENTER_CAT,
                                                       SINK_CAT,
                                                       SUBSTATION_CAT,
                                                       MINIMUM_SPUR_DIST_KM,
                                                       BARRIERS_MULT,
                                                       CLIP_RASTER_BUFFER,
                                                       NUM_NN_SINKS,
                                                       BARRIER_H5_LAYER_NAME,
                                                       ISO_H5_LAYER_NAME)
from reVX.least_cost_xmission.least_cost_paths import min_reinforcement_costs

TRANS_CAT_TYPES = [TRANS_LINE_CAT, LOAD_CENTER_CAT, SINK_CAT, SUBSTATION_CAT]

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default='LeastCostXmission', type=STR,
              show_default=True,
              help='Job name.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """
    Least Cost Xmission Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose
    ctx.obj['NAME'] = name


@main.command()
def valid_config_keys():
    """
    Echo the valid Least Cost Xmission config keys
    """
    click.echo(', '.join(get_class_properties(LeastCostXmissionConfig)))


@main.command()
@click.option('--config', '-c', required=True,
              type=click.Path(exists=True),
              help='Filepath to Least Cost Xmission config json file.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config, verbose):
    """
    Run Least Cost Xmission from a config.
    """

    config = LeastCostXmissionConfig(config)
    option = config.execution_control.option

    if 'VERBOSE' in ctx.obj:
        if any((ctx.obj['VERBOSE'], verbose)):
            config._log_level = logging.DEBUG
    elif verbose:
        config._log_level = logging.DEBUG

    if option == 'local':
        run_local(ctx, config)
        return

    if option not in {'eagle', 'kestrel'}:
        click.echo('Option "{}" is not supported'.format(option))
        return

    if config.execution_control.nodes == 1:
        eagle(config, config.sc_point_gids)
        return

    # Split gids over mulitple SLURM jobs
    name = config.name
    logger.info('Splitting SC points over {} SLURM jobs'
                .format(config.execution_control.nodes))
    for i in range(config.execution_control.nodes):
        config.name = '{}_{}'.format(name, i)
        eagle(config, config.sc_point_gids[i::config.execution_control.nodes])


@main.command()
@click.option('--cost_fpath', '-cost', type=click.Path(exists=True),
              required=True,
              help=("Path to h5 file with cost rasters and other required "
                    "layers"))
@click.option('--features_fpath', '-feats', required=True,
              type=click.Path(exists=True),
              help="Path to GeoPackage with transmission features")
@click.option('--regions_fpath', '-regs', type=STR, show_default=True,
              default=None,
              help=("Path to reinforcement regions GeoPackage. If not `None`, "
                    "Least Cost Xmission is run with reinforcement path "
                    "costs. Features must be substations only, and the "
                    "substation file must contain a "
                    "`region_identifier_column` column that matches the "
                    "`region_identifier_column` ID in this file for the "
                    "reinforcement region containing that substation. "))
@click.option('--region_identifier_column', '-rid', type=STR, default=None,
              help=("Name of column in reinforcement regions GeoPackage"
                    "containing a unique identifier for each region."))
@click.option('--capacity_class', '-cap', type=str, required=True,
              help=("Capacity class of transmission features to connect "
                    "supply curve points to"))
@click.option('--resolution', '-res', type=int,
              show_default=True, default=128,
              help=("SC point resolution"))
@click.option('--xmission_config', '-xcfg', type=STR, show_default=True,
              default=None,
              help=("Path to Xmission config .json"))
@click.option('--min_line_length', '-mll', type=int,
              show_default=True, default=MINIMUM_SPUR_DIST_KM,
              help=("Minimum Tie-line length."))
@click.option('--sc_point_gids', '-gids', type=INTLIST, show_default=True,
              default=None, help=("List of sc_point_gids to connect to. If "
                                  "running `from_config`, this can also be a "
                                  "path to a CSV file with a 'sc_point_gid' "
                                  "column containing the GID's to run. Note "
                                  "the missing 's' in the column name - this "
                                  "makes it seamless to run on a supply curve "
                                  "output from reV"))
@click.option('--nn_sinks', '-nn', type=int,
              show_default=True, default=NUM_NN_SINKS,
              help=("Number of nearest neighbor sinks to use for clipping "
                    "radius calculation. This is overridden by --radius"))
@click.option('--clipping_buffer', '-buffer', type=float,
              show_default=True, default=CLIP_RASTER_BUFFER,
              help=("Buffer to expand clipping radius by"))
@click.option('--tb_layer_name', '-tbln', default=BARRIER_H5_LAYER_NAME,
              type=STR, show_default=True,
              help='Name of transmission barrier layer in `cost_fpath` file. '
                   'This layer defines the multipliers applied to the cost '
                   'layer to determine LCP routing (but does not actually '
                   'affect output costs')
@click.option('--barrier_mult', '-bmult', type=float,
              show_default=True, default=BARRIERS_MULT,
              help=("Tranmission barrier multiplier, used when computing the "
                    "least cost tie-line path"))
@click.option('--iso_regions_layer_name', '-irln', default=ISO_H5_LAYER_NAME,
              type=STR, show_default=True,
              help='Name of ISO regions layer in `cost_fpath` file. The '
                   "layer maps pixels to ISO region ID's (1, 2, 3, 4, etc.) .")
@click.option('--max_workers', '-mw', type=INT,
              show_default=True, default=None,
              help=("Number of workers to use for processing, if 1 run in "
                    "serial, if None use all available cores"))
@click.option('--out_dir', '-o', type=STR, default='./out',
              show_default=True,
              help='Directory to save least cost xmission values to.')
@click.option('--log_dir', '-log', default=None, type=STR,
              show_default=True,
              help='Directory to dump log files.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.option('--save_paths', is_flag=True,
              help='Save least cost paths and data to GeoPackage.')
@click.option('--radius', '-rad', type=INT,
              show_default=True, default=None,
              help=("Radius to clip costs raster to in pixels This overrides "
                    "--nn_sinks if set."))
@click.option('--mp_delay', '-mpd', type=FLOAT, show_default=True, default=3.0,
              help=("Delay in seconds between starting multi-process workers. "
                    "Useful for reducing memory spike at working startup."))
@click.option('--expand_radius', '-er', is_flag=True,
              help='Flag to expand radius until at least one transmission '
                   'feature is included for connection. Has no effect if '
                   'radius input is ``None``.')
@click.option('--simplify-geo', type=FLOAT,
              show_default=True, default=None,
              help=("Simplify path geometries by a value before writing to "
                    "GeoPackage."))
@click.option('--cost-layers', '-cl', required=True, multiple=True,
              default=(),
              help='Layer in H5 to add to total cost raster used for routing. '
                   'Multiple layers may be specified. Layer name may have '
                   'curly brackets (``{}``), which will be filled in '
                   'based on the capacity class input (e.g. '
                   '"tie_line_costs_{}MW")')
@click.option('--li-cost-layers', '-licl', required=False, multiple=True,
              default=(),
              help='Length-invariant cost layer in H5 to add to total cost '
                   'raster used for routing. These costs do not scale with '
                   'distance traversed acroiss the cell. Multiple layers may '
                   'be specified.')
@click.option('--tracked_layers', '-trl', type=STR, default=None,
              show_default=True,
              help=('Dictionary mapping layer names to strings, where the '
                    'strings are numpy methods that should be applied to '
                    'the layer along the LCP.'))
@click.option('--length_mult_kind', '-lmk', type=STR, default='linear',
              show_default=True,
              help='Type of length multiplier calcualtion. "step" computes '
                   'length multipliers using a step function, while "linear" '
                   'computes the length multiplier using a linear '
                   'interpolation between 0 amd 10 mile spur-line lengths.')
@click.option('--cell_size', '-cs', type=INT,
              show_default=True, default=CELL_SIZE,
              help=("Side length of a single cell in meters. Cells are "
                    "assumed to be square. Default is"))
@click.pass_context
def local(ctx, cost_fpath, features_fpath, regions_fpath,
          region_identifier_column, capacity_class, resolution,
          xmission_config, min_line_length, sc_point_gids, nn_sinks,
          clipping_buffer, tb_layer_name, barrier_mult,
          iso_regions_layer_name, max_workers, out_dir, log_dir, verbose,
          save_paths, radius, expand_radius, mp_delay, simplify_geo,
          cost_layers: List[str], li_cost_layers, tracked_layers,
          length_mult_kind, cell_size):
    """
    Run Least Cost Xmission on local hardware
    """
    start_time = time.time()
    name = ctx.obj['NAME']
    if 'VERBOSE' in ctx.obj:
        verbose = any((ctx.obj['VERBOSE'], verbose))

    log_modules = [__name__, 'reVX', 'reV', 'rex']
    init_mult(name, log_dir, modules=log_modules, verbose=verbose)

    create_dirs(out_dir)
    logger.info('Computing Least Cost Xmission connections and writing them {}'
                .format(out_dir))

    cost_layers = list(cost_layers)
    li_cost_layers = list(li_cost_layers)
    if isinstance(tracked_layers, str):
        tracked_layers = dict_str_load(tracked_layers)

    kwargs = {"resolution": resolution,
              "xmission_config": xmission_config,
              "min_line_length": min_line_length,
              "sc_point_gids": sc_point_gids,
              "clipping_buffer": clipping_buffer,
              "tb_layer_name": tb_layer_name,
              "barrier_mult": barrier_mult,
              "iso_regions_layer_name": iso_regions_layer_name,
              "max_workers": max_workers,
              "save_paths": save_paths,
              "simplify_geo": simplify_geo,
              "radius": radius,
              "expand_radius": expand_radius,
              "mp_delay": mp_delay,
              "length_invariant_cost_layers": li_cost_layers,
              "length_mult_kind": length_mult_kind,
              "tracked_layers": tracked_layers,
              "cell_size": cell_size}

    if regions_fpath is not None:
        least_costs = RegionalXmission.run(cost_fpath, features_fpath,
                                           regions_fpath,
                                           region_identifier_column,
                                           capacity_class, cost_layers,
                                           **kwargs)
    else:
        kwargs["nn_sinks"] = nn_sinks
        least_costs = LeastCostXmission.run(cost_fpath, features_fpath,
                                            capacity_class, cost_layers,
                                            **kwargs)
    if len(least_costs) == 0:
        logger.error('No paths found.')
        return

    ext = 'gpkg' if save_paths else 'csv'
    fn_out = '{}_{}_{}.{}'.format(name, capacity_class, resolution, ext)
    fpath_out = os.path.join(out_dir, fn_out)

    logger.info('Writing output to %s', fpath_out)
    if save_paths:
        least_costs.to_file(fpath_out, driver='GPKG')
    else:
        least_costs.to_csv(fpath_out, index=False)

    logger.info('Writing output complete')
    elapsed_time = (time.time() - start_time) / 60
    logger.info("Processing took %.2f minutes", elapsed_time)


@main.command()
@click.option('--split-to-geojson', '-s', is_flag=True,
              help='After merging GeoPackages, split into GeoJSON by POI name'
              '.')
@click.option('--suppress-combined-file', is_flag=True,
              help='Don\'t create combined layer.')
@click.option('--out-file', '-of', default=None, type=STR,
              help='Name for output GeoPackage/CSV file.')
@click.option('--drop', '-d', default=None, type=STR, multiple=True,
              help=('Transmission feature category types to drop from '
                    'results. Options: {}'.format(", ".join(TRANS_CAT_TYPES))))
@click.option('--out-dir', '-od', type=click.Path(),
              default='./out', show_default=True,
              help='Output directory for output files.')
@click.option('--simplify-geo', type=FLOAT,
              show_default=True, default=None,
              help='Simplify path geometries by a value before exporting.')
@click.option('--ss_id_col', '-ssid', default=None, type=STR,
              show_default=True,
              help='Name of column used to unqiuely identify substations. '
                   'Used for reinforcement calcaultions only. ')
@click.argument('files', type=STR, nargs=-1)
@click.pass_context
# flake8: noqa: C901
def merge_output(ctx, split_to_geojson, suppress_combined_file, out_file,
                 out_dir, drop, simplify_geo, ss_id_col, files):
    """
    Merge output GeoPackage/CSV files and optionally convert to GeoJSON
    """
    log_level = "DEBUG" if ctx.obj.get('VERBOSE') else "INFO"
    init_logger('reVX', log_level=log_level)

    if len(files) == 0:
        logger.error('No files passed to be merged')
        return

    if drop:
        for cat in drop:
            if cat not in TRANS_CAT_TYPES:
                logger.info('--drop options must be one or more of %s, '
                            'received %s', TRANS_CAT_TYPES, drop)
                return

    warnings.filterwarnings('ignore', category=RuntimeWarning)
    dfs = []
    for i, file in enumerate(files, start=1):
        logger.info('Loading %s (%i/%i)', file, i, len(files))
        df_tmp = gpd.read_file(file) if "gpkg" in file else pd.read_csv(file)
        dfs.append(df_tmp)

    df = pd.concat(dfs)
    warnings.filterwarnings('default', category=RuntimeWarning)

    if drop:
        mask = df['category'].isin(drop)
        logger.info('Dropping %d of %d total features with category(ies): %s',
                    mask.sum(), len(df), ", ".join(drop))
        df = df[~mask]

    df = df.reset_index()

    if len(df) == 0:
        logger.info('No transmission features to save.')
        return

    if simplify_geo:
        logger.info('Simplifying geometries by {}'.format(simplify_geo))
        df.geometry = df.geometry.simplify(simplify_geo)

    if ss_id_col is not None:
        logger.info('Computing minimum reinforcement cost')
        df = min_reinforcement_costs(df, group_col=ss_id_col)

    create_dirs(out_dir)

    # Create combined output file
    if not suppress_combined_file:
        out_file = ('combo_{}'.format(files[0])
                    if out_file is None else out_file)
        out_file = os.path.join(out_dir, out_file)
        logger.info('Saving all combined paths to %s', out_file)
        if "gpkg" in out_file:
            df.to_file(out_file, driver="GPKG")
        else:
            df.to_csv(out_file, index=False)

    # Split out put in to GeoJSON by POI name
    if split_to_geojson:
        if not isinstance(df, gpd.GeoDataFrame):
            click.echo('Geo-spatial aware input files must be provided to '
                       'split to Geo-JSON.')
            sys.exit(1)
        pois = set(df['POI Name'])
        for i, poi in enumerate(pois, start=1):
            safe_poi_name = poi.replace(' ', '_').replace('/', '_')
            out_file = os.path.join(out_dir, f"{safe_poi_name}_paths.geojson")
            paths = df[df['POI Name'] == poi].to_crs(epsg=4326)
            logger.info('Writing {} paths for {} to {} ({}/{})'
                        .format(len(paths), poi, out_file, i, len(pois)))
            paths.to_file(out_file, driver="GeoJSON")


@main.command()
@click.option('--cost_fpath', '-f', required=True,
              type=click.Path(exists=True),
              help=("Path to GeoPackage/CSV file with calculated transmission "
                    "costs. This file must have a 'trans_gid' column that "
                    "will be used to merge in the reinforcement costs."))
@click.option('--reinforcement_cost_fpath', '-r', required=True,
              type=click.Path(exists=True),
              help=("Path to GeoPackage/CSV file with calculated "
                    "reinforcement costs. This file must have a 'gid' column "
                    "that will be used to merge in the reinforcement costs."))
@click.option('--merge_column', '-mc', default="trans_gid", type=STR,
              help=("Name of column in `cost_fpath` and "
                    "`reinforcement_cost_fpath` files to merge on."))
@click.option('--out_file', '-of', default=None, type=STR,
              help='Name for output GeoPackage/CSV file.')
@click.pass_context
def merge_reinforcement_costs(ctx, cost_fpath, reinforcement_cost_fpath,
                              merge_column, out_file):
    """
    Merge reinforcement costs into transmission costs.
    """
    log_level = "DEBUG" if ctx.obj.get('VERBOSE') else "INFO"
    init_logger('reVX', log_level=log_level)
    logger.info("Merging reinforcement costs into transmission costs...")

    logger.debug("Reading in transmission costs from %s", str(cost_fpath))
    costs = (gpd.read_file(cost_fpath)
             if "gpkg" in cost_fpath
             else pd.read_csv(cost_fpath))

    logger.debug("Reading in reinforcement costs from %s",
                 str(reinforcement_cost_fpath))
    r_costs = (gpd.read_file(reinforcement_cost_fpath)
               if "gpkg" in reinforcement_cost_fpath
               else pd.read_csv(reinforcement_cost_fpath))

    logger.info("Loaded spur-line costs for %d substations and "
                "reinforcement costs for %d substations",
                len(costs[merge_column].unique()),
                len(r_costs[merge_column].unique()))

    r_costs.index = r_costs[merge_column].values
    costs = costs[costs[merge_column].isin(r_costs[merge_column])].copy()

    logger.info("Found %d substations with both spur-line and "
                "reinforcement costs",
                len(costs[merge_column].unique()))

    r_cols = ["reinforcement_poi_lat", "reinforcement_poi_lon",
              "reinforcement_dist_km", "reinforcement_cost_per_mw"]
    costs[r_cols] = r_costs.loc[costs[merge_column], r_cols].values

    logger.info("Writing output to {!r}".format(out_file))

    if "gpkg" in out_file:
        costs.to_file(out_file, driver="GPKG", index=False)
    else:
        costs.to_csv(out_file, index=False)


def get_node_cmd(config, gids):
    """
    Get the node CLI call for Least Cost Xmission

    Parameters
    ----------
    config : reVX.config.least_cost_xmission.LeastCostXmissionConfig
        Least Cost Xmission config object.
    gids : list
        List of SC point GID values to submit to local command.

    Returns
    -------
    cmd : str
        CLI call to submit to SLURM execution.
    """
    args = ['-n {}'.format(SLURM.s(config.name)),
            'local',
            '-cost {}'.format(SLURM.s(config.cost_fpath)),
            '-feats {}'.format(SLURM.s(config.features_fpath)),
            '-regs {}'.format(SLURM.s(config.regions_fpath)),
            '-rid {}'.format(SLURM.s(config.region_identifier_column)),
            '-cap {}'.format(SLURM.s(config.capacity_class)),
            '-res {}'.format(SLURM.s(config.resolution)),
            '-xcfg {}'.format(SLURM.s(config.xmission_config)),
            '-mll {}'.format(SLURM.s(config.min_line_length)),
            '-gids {}'.format(SLURM.s(gids)),
            '-nn {}'.format(SLURM.s(config.nn_sinks)),
            '-buffer {}'.format(SLURM.s(config.clipping_buffer)),
            '-tbln {}'.format(SLURM.s(config.tb_layer_name)),
            '-bmult {}'.format(SLURM.s(config.barrier_mult)),
            '-irln {}'.format(SLURM.s(config.iso_regions_layer_name)),
            '-mw {}'.format(SLURM.s(config.execution_control.max_workers)),
            '-o {}'.format(SLURM.s(config.dirout)),
            '-log {}'.format(SLURM.s(config.log_directory)),
            '-lmk {}'.format(SLURM.s(config.length_mult_kind)),
            '-mpd {}'.format(SLURM.s(config.mp_delay)),
            '-cs {}'.format(SLURM.s(config.cell_size)),
            ]

    for layer in config.cost_layers:
        args.append(f'-cl {layer}')
    for layer in config.length_invariant_cost_layers:
        args.append(f'-licl {layer}')

    if config.save_paths:
        args.append('--save_paths')
    if config.radius:
        args.append('-rad {}'.format(config.radius))
    if config.expand_radius:
        args.append('-er')
    if config.simplify_geo:
        args.append('--simplify-geo {}'.format(config.simplify_geo))
    if config.tracked_layers:
        args.append('-trl {}'.format(SLURM.s(config.tracked_layers)))

    if config.log_level == logging.DEBUG:
        args.append('-v')

    cmd = ('python -m reVX.least_cost_xmission.least_cost_xmission_cli {}'
           .format(' '.join(args)))
    logger.debug('Submitting the following cli call:\n\t{}'.format(cmd))

    return cmd


def run_local(ctx, config):
    """
    Run Least Cost Xmission locally using config

    Parameters
    ----------
    ctx : click.ctx
        click ctx object
    config : reVX.config.least_cost_xmission.LeastCostXmissionConfig
        Least Cost Xmission config object.
    """
    ctx.obj['NAME'] = config.name
    ctx.invoke(local,
               cost_fpath=config.cost_fpath,
               features_fpath=config.features_fpath,
               regions_fpath=config.regions_fpath,
               capacity_class=config.capacity_class,
               resolution=config.resolution,
               xmission_config=config.xmission_config,
               min_line_length=config.min_line_length,
               sc_point_gids=config.sc_point_gids,
               nn_sinks=config.nn_sinks,
               clipping_buffer=config.clipping_buffer,
               tb_layer_name=config.tb_layer_name,
               barrier_mult=config.barrier_mult,
               iso_regions_layer_name=config.iso_regions_layer_name,
               region_identifier_column=config.region_identifier_column,
               max_workers=config.execution_control.max_workers,
               out_dir=config.dirout,
               log_dir=config.log_directory,
               verbose=config.log_level,
               radius=config.radius,
               expand_radius=config.expand_radius,
               mp_delay=config.mp_delay,
               save_paths=config.save_paths,
               simplify_geo=config.simplify_geo,
               cost_layers=config.cost_layers,
               li_cost_layers=config.length_invariant_cost_layers,
               tracked_layers=config.tracked_layers,
               length_mult_kind=config.length_mult_kind,
               cell_size=config.cell_size)


def eagle(config, gids):
    """
    Run Least Cost Xmission on Eagle HPC.

    Parameters
    ----------
    config : reVX.config.least_cost_xmission.LeastCostXmissionConfig
        Least Cost Xmission config object.
    """
    init_logger('rex', log_level='DEBUG')

    cmd = get_node_cmd(config, gids)
    name = config.name
    log_dir = config.log_directory
    stdout_path = os.path.join(log_dir, 'stdout/')

    slurm_manager = SLURM()

    logger.info('Running Least Cost Xmission on Eagle with '
                'node name "{}"'.format(name))
    out = slurm_manager.sbatch(cmd,
                               alloc=config.execution_control.allocation,
                               memory=config.execution_control.memory,
                               walltime=config.execution_control.walltime,
                               feature=config.execution_control.feature,
                               name=name, stdout_path=stdout_path,
                               conda_env=config.execution_control.conda_env,
                               module=config.execution_control.module)[0]
    if out:
        msg = ('Kicked off Least Cost Xmission "{}" '
               '(SLURM jobid #{}) on Eagle.'
               .format(name, out))
    else:
        msg = ('Was unable to kick off Least Cost Xmission '
               '"{}". Please see the stdout error messages'
               .format(name))

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running Least Cost Xmission CLI')
        raise



# -*- coding: utf-8 -*-
"""
reVX Least Cost Xmission Configurations
"""
import os
import logging
import pandas as pd
from typing import List

from reVX.least_cost_xmission.config.constants import (CELL_SIZE,
                                                       RESOLUTION,
                                                       NUM_NN_SINKS,
                                                       CLIP_RASTER_BUFFER,
                                                       BARRIERS_MULT,
                                                       MINIMUM_SPUR_DIST_KM,
                                                       BARRIER_H5_LAYER_NAME,
                                                       ISO_H5_LAYER_NAME)
from reV.supply_curve.extent import SupplyCurveExtent
from reV.config.base_analysis_config import AnalysisConfig
from reV.utilities.exceptions import ConfigError

logger = logging.getLogger(__name__)


class LeastCostXmissionConfig(AnalysisConfig):
    """Config framework for Least Cost Xmission"""

    NAME = 'LeastCostXmission'
    REQUIREMENTS = ('cost_fpath', 'features_fpath', 'capacity_class',
                    'cost_layers')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            Path to config .json or pre-extracted config input dictionary.
        """
        super().__init__(config)
        self._sc_point_gids = None

    @property
    def name(self):
        """Get the job name, defaults to the output directory name.

        Returns
        -------
        _name : str
            Job name.
        """
        if self._name is None:
            default_name = os.path.basename(os.path.normpath(self.dirout))
            self._name = self.get('name', default_name)
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def radius(self):
        """
        Optional radius to use for clipping
        """
        return self.get('radius', None)

    @property
    def expand_radius(self):
        """
        Optional flag to expand radius until at least one transmission
        feature is available for connection.
        """
        return self.get('expand_radius', True)

    @property
    def mp_delay(self):
        """
        Delay in seconds between starting multi-process workers.
        Useful for reducing memory spike at working startup.
        By default, ``3.0``.
        """
        return self.get('mp_delay', 3.0)

    @property
    def simplify_geo(self):
        """
        Optional float to use to simplify path geometries before saving to
        geopackage
        """
        return self.get('simplify_geo', None)

    @property
    def save_paths(self):
        """
        Save paths as GPKG if true
        """
        return self.get('save_paths', False)

    @property
    def cost_fpath(self):
        """
        .h5 file with cost and required layers
        """
        return self['cost_fpath']

    @property
    def features_fpath(self):
        """
        Tranmission feature .gpkg
        """
        return self['features_fpath']

    @property
    def regions_fpath(self):
        """
        Reinforcement regions .gpkg
        """
        return self.get('regions_fpath', None)

    @property
    def region_identifier_column(self):
        """
        Name of column containing unique region identifier
        """
        rid_col = self.get("region_identifier_column", None)
        if self.regions_fpath is not None and rid_col is None:
            msg = ("`region_identifier_column` input cannot be `None` for "
                   "reinforcement path computation.")
            raise ConfigError(msg)
        return rid_col

    @property
    def capacity_class(self):
        """
        Capacity class, either {capacity}MW or capacity value in MW
        """
        return self['capacity_class']

    @property
    def resolution(self):
        """
        SC point resolution
        """
        return self.get('resolution', RESOLUTION)

    @property
    def xmission_config(self):
        """
        Xmission config input
        """
        return self.get('xmission_config', None)

    @property
    def length_mult_kind(self):
        """
        Type of length multiplier calcualtion.

        "step" computes length multipliers using a step function, while
        "linear" computes the length multiplier using a linear
        interpolation between 0 amd 10 mile spur-line lengths.
        By default, ``"linear"``.
        """
        return self.get('length_mult_kind', "linear")

    @property
    def min_line_length(self):
        """
        Minimum Tie-line length config input
        """
        return self.get('min_line_length', MINIMUM_SPUR_DIST_KM)

    @property
    def nn_sinks(self):
        """
        Number of nearest neighbor sinks to use for clipping radius
        """
        return self.get('nn_sinks', NUM_NN_SINKS)

    @property
    def clipping_buffer(self):
        """
        Buffer to add to clipping radius
        """
        return self.get('clipping_buffer', CLIP_RASTER_BUFFER)

    @property
    def tb_layer_name(self):
        """
        Name of transmission barrier layer in `cost_fpath` file.
        """
        return self.get('tb_layer_name', BARRIER_H5_LAYER_NAME)

    @property
    def barrier_mult(self):
        """
        Transmission barrier multiplier to use for MCP costs
        """
        return float(self.get('barrier_mult', BARRIERS_MULT))

    @property
    def iso_regions_layer_name(self):
        """
        Name of ISO regions layer in `cost_fpath` file.
        """
        return self.get('iso_regions_layer_name', ISO_H5_LAYER_NAME)

    @property
    def cost_layers(self) -> List[str]:
        """
        List of H5 layers that are summed to determine total costs
        raster used for routing. Costs and distances for each individual
        layer are also reported (e.g. wet and dry costs).
        """
        return self['cost_layers']

    @property
    def length_invariant_cost_layers(self):
        """
        Layers to be added to the cost raster whose costs do not scale
        with distance traversed (i.e. fixed one-time costs for crossing
        these cells).
        """
        # self.get('length_invariant_cost_layers', []) does not work!!
        if 'length_invariant_cost_layers' not in self:
            return []
        return self['length_invariant_cost_layers']

    @property
    def tracked_layers(self):
        """
        Dictionary mapping layer names to strings, where the strings are
        numpy methods that should be applied to the layer along the LCP.
        """
        return self.get("tracked_layers")

    @property
    def cell_size(self):
        """
        Side length of each cell, in meters.
        """
        return self.get('cell_size', CELL_SIZE)

    @property
    def sc_point_gids(self):
        """
        List of sc_point_gids to compute Least Cost Xmission for
        """
        if self._sc_point_gids is None:
            sc_point_gids = self.get('sc_point_gids')
            if sc_point_gids is None:
                sce = SupplyCurveExtent(self.cost_fpath,
                                        resolution=self.resolution)
                sc_point_gids = list(sce.points.index.values)
            elif (isinstance(sc_point_gids, str)
                  and sc_point_gids.endswith(".csv")):
                points = pd.read_csv(sc_point_gids)
                sc_point_gids = list(points.sc_point_gid.values)

            if not isinstance(sc_point_gids, list):
                raise ValueError('sc_point_gids must be a list or path to a '
                                 'csv file, got a {} ({})'
                                 .format(type(sc_point_gids), sc_point_gids))
            self._sc_point_gids = sc_point_gids

        return self._sc_point_gids


class LeastCostPathsConfig(AnalysisConfig):
    """Config framework for Least Cost Paths"""

    NAME = 'LeastCostPaths'
    REQUIREMENTS = ('cost_fpath', 'features_fpath', 'cost_layers')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : str | dict
            Path to config .json or pre-extracted config input dictionary.
        """
        super().__init__(config)
        self._validate_reinforcement_run()

    def _validate_reinforcement_run(self):
        """Validate all inputs given for reinforcement run"""
        reinforcement_run = any(val is not None for val in
                                [self.transmission_lines_fpath,
                                 self.network_nodes_fpath,
                                 self.capacity_class])
        missing_inputs = any(val is None for val in
                             [self.transmission_lines_fpath,
                              self.network_nodes_fpath,
                              self.capacity_class])
        if reinforcement_run and missing_inputs:
            msg = ("Must specify all of the following arguments for "
                   "reinforcement computations: 'transmission_lines_fpath', "
                   "'network_nodes_fpath', and 'capacity_class'!")
            logger.error(msg)
            raise ValueError(msg)

    @property
    def name(self):
        """Get the job name, defaults to the output directory name.

        Returns
        -------
        _name : str
            Job name.
        """
        if self._name is None:
            default_name = os.path.basename(os.path.normpath(self.dirout))
            self._name = self.get('name', default_name)
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def cost_fpath(self):
        """
        .h5 file with cost and required layers
        """
        return self['cost_fpath']

    @property
    def features_fpath(self):
        """
        Tranmission feature .gpkg
        """
        return self['features_fpath']

    @property
    def network_nodes_fpath(self):
        """
        Network nodes config input
        """
        return self.get('network_nodes_fpath', None)

    @property
    def transmission_lines_fpath(self):
        """
        Transmission line features  config input
        """
        return self.get('transmission_lines_fpath', None)

    @property
    def capacity_class(self):
        """
        Capacity class of the 'base' greenfield costs layer. Costs will
        be scaled by the capacity corresponding to this class to report
        reinforcement costs as $/MW. Only used for reinforcement path
        computations.
        """
        return self.get("capacity_class")

    @property
    def cost_layers(self):
        """
        List of H5 layers that are summed to determine total costs
        raster used for routing.
        """
        return self['cost_layers']

    @property
    def length_invariant_cost_layers(self):
        """
        Layers to be added to the cost raster whose costs do not scale
        with distance traversed (i.e. fixed one-time costs for crossing
        these cells).
        """
        # self.get('length_invariant_cost_layers', []) does not work!!
        if 'length_invariant_cost_layers' not in self:
            return []
        return self['length_invariant_cost_layers']

    @property
    def tracked_layers(self):
        """
        Dictionary mapping layer names to strings, where the strings are
        numpy methods that should be applied to the layer along the LCP.
        """
        return self.get("tracked_layers")

    @property
    def cell_size(self):
        """
        Side length of each cell, in meters.
        """
        return self.get('cell_size', CELL_SIZE)

    @property
    def xmission_config(self):
        """
        Xmission config input
        """
        return self.get('xmission_config', None)

    @property
    def tb_layer_name(self):
        """
        Name of transmission barrier layer in `cost_fpath` file.
        """
        return self.get('tb_layer_name', BARRIER_H5_LAYER_NAME)

    @property
    def clip_buffer(self):
        """
        Number of array elements to buffer clip area by.
        """
        return self.get('clip_buffer', 0)

    @property
    def barrier_mult(self):
        """
        Transmission barrier multiplier to use for MCP costs
        """
        return self.get('barrier_mult', BARRIERS_MULT)

    @property
    def is_reinforcement_run(self):
        """
        Boolean flag indicating wether this run is for reinforcement
        path computation
        """
        return (self.network_nodes_fpath is not None
                and self.transmission_lines_fpath is not None)

    @property
    def region_identifier_column(self):
        """
        Name of column containing unique region identifier
        """
        rid_col = self.get("region_identifier_column", None)
        if self.is_reinforcement_run and rid_col is None:
            msg = ("`region_identifier_column` input cannot be `None` for "
                   "reinforcement path computation.")
            raise ConfigError(msg)
        return rid_col

    @property
    def ss_id_col(self):
        """
        Name of column containing unique identifier for each substation.
        This column will be used to compute minimum reinforcement cost
        per substation. By default, ``"poi_gid"``.
        """
        return self.get("ss_id_col", "poi_gid")

    @property
    def save_paths(self):
        """
        Boolean flag to save the least cost paths along with the costs and
        lengths
        """
        return self.get('save_paths', False)



# -*- coding: utf-8 -*-
"""Handler for H5 file containing GeoTIFF layers. """
import os
import json
import logging
from pathlib import Path
from copy import deepcopy
from warnings import warn

import h5py
import numpy as np
from pyproj.crs import CRS
from rasterio.warp import reproject, Resampling

from rex import Resource, Outputs
from reVX.handlers.geotiff import Geotiff
from reVX.utilities.exceptions import ProfileCheckError


logger = logging.getLogger(__name__)


class LayeredH5:
    """Handler for HDF5 file containing GeTIFF layers.

    This handler represents an HDF5 file that stores various layers
    (i.e. exclusion layers, setback layers, transmission layers, etc).
    This file contains profile information, and this handler can be used
    to convert to and from such files.
    """

    SUPPORTED_FILE_ENDINGS = {".h5", ".tif", ".tiff"}
    """Supported template file endings."""

    LATITUDE = "latitude"
    """Name of latitude values layer in HDF5 file."""

    LONGITUDE = "longitude"
    """Name of longitude values layer in HDF5 file."""

    def __init__(self, h5_file, hsds=False, chunks=(128, 128),
                 template_file=None, block_size=None):
        """

        Parameters
        ----------
        h5_file : path-like
            Path to HDF5 layered file. If this file is to be created,
            a `template_file` must be provided (and must exist on disk).
            Otherwise, the `template_file` input can be ignored and this
            input will be used as the template file.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle HDF5 'files' hosted on
            AWS behind HSDS. By default, ``False``.
        chunks : tuple, optional
            Chunk size of exclusions in HDF5 file and any output
            GeoTIFFs. By default, ``(128, 128)``.
        template_file : path-like, optional
            Path to template GeoTIFF (``*.tif`` or ``*.tiff``) or HDF5
            (``*.h5``) file containing the profile and transform to be
            used for the layered file. If ``None``, then the `h5_file`
            input is used as the template. By default, ``None``.
        block_size : int, optional
            Optional block size to use when building lat/lon datasets.
            Setting this value can help reduce memory issues when
            building a ``LayeredH5`` file. If ``None``, the lat/lon
            arrays are processed in full. By default, ``None``.
        """
        self.h5_file = h5_file
        self._hsds = hsds
        self._chunks = chunks
        self._profile = None
        self._template_file = template_file or h5_file
        self._block_size = block_size

    def __repr__(self):
        return "{} for {}".format(self.__class__.__name__, self.h5_file)

    def __setitem__(self, layer_name, geotiff):
        self.write_geotiff_to_h5(geotiff, layer_name)

    def __getitem__(self, layer):
        if layer not in self.layers:
            msg = "{} is not present in {}".format(layer, self.h5_file)
            logger.error(msg)
            raise KeyError(msg)

        logger.debug('\t- Extracting %s from %s', layer, self.h5_file)
        with Resource(self.h5_file, hsds=self._hsds) as h5:
            profile = h5.get_attrs(dset=layer).get('profile')
            values = h5[layer]

        if profile is not None:
            profile = json.loads(profile)

        return profile, values

    def _validate_template(self):
        """Validate template file. """
        valid_file_ending = any(str(self.template_file).endswith(fe)
                                for fe in self.SUPPORTED_FILE_ENDINGS)
        if not valid_file_ending:
            msg = ("Template file {!r} format is not supported! File must end "
                   "in one of: {}".format(self.template_file,
                                          self.SUPPORTED_FILE_ENDINGS))
            logger.error(msg)
            raise ValueError(msg)

        if not self._hsds and not Path(self.template_file).exists():
            msg = ("Template file {!r} not found on disk!"
                   .format(self.template_file))
            logger.error(msg)
            raise FileNotFoundError(msg)

    def _extract_profile(self):
        """Extract template profile. """
        if str(self.template_file).endswith(".h5"):
            with Resource(self.template_file, hsds=self._hsds) as h5:
                return json.loads(h5.global_attrs['profile'])

        with Geotiff(self.template_file) as geo:
            return geo.profile

    # pylint: disable=unpacking-non-sequence
    def _extract_lat_lon(self):
        """Extract template lat/lons. """
        if str(self.template_file).endswith(".h5"):
            with Resource(self.template_file, hsds=self._hsds) as h5:
                return h5[self.LATITUDE], h5[self.LONGITUDE]

        with Geotiff(self.template_file) as geo:
            if not self._block_size:
                return geo.lat_lon

            nrows, ncols = geo.shape
            out_lat = np.zeros((nrows, ncols), dtype="float32")
            out_lon = np.zeros((nrows, ncols), dtype="float32")
            for x in range(0, nrows, self._block_size):
                for y in range(0, ncols, self._block_size):
                    logger.debug("Loading lat/lon starting at inds %d, %d",
                                 x, y)
                    r_slice = slice(x, x + self._block_size)
                    c_slice = slice(y, y + self._block_size)
                    lat, lon = geo["lat_lon", r_slice, c_slice]
                    out_lat[r_slice, c_slice] = lat
                    out_lon[r_slice, c_slice] = lon

        return out_lat, out_lon

    @property
    def template_file(self):
        """str: Path to template file. """
        return self._template_file

    @template_file.setter
    def template_file(self, new_template_file):
        self._template_file = new_template_file
        self._validate_template()

    @property
    def profile(self):
        """dict: Template layer profile. """
        if self._profile is None:
            self._profile = self._extract_profile()
        return self._profile

    @property
    def shape(self):
        """tuple: Template layer shape. """
        return self.profile['height'], self.profile['width']

    @property
    def layers(self):
        """list: Available layers in HDF5 file. """
        if not Path(self.h5_file).exists():
            msg = f"File {self.h5_file!r} not found"
            logger.error(msg)
            raise FileNotFoundError(msg)

        with Resource(self.h5_file) as h5:
            return h5.datasets

    def create_new(self, overwrite=False):
        """Create a new layered HDF5 file.

        Parameters
        ----------
        overwrite : bool, optional
            Overwrite HDF5 file if is exists. By default, ``False``.
        """
        if Path(self.h5_file).exists() and not overwrite:
            msg = f"File {self.h5_file!r} exits and overwrite=False"
            logger.error(msg)
            raise FileExistsError(msg)

        if self.h5_file == self.template_file:
            msg = f"Must provide template file to initialize {self.h5_file}!"
            logger.error(msg)
            raise ValueError(msg)

        self._validate_template()

        logger.debug('\t- Initializing %s from %s',
                     self.h5_file, self.template_file)

        lat, lon = self._extract_lat_lon()
        logger.debug('\t- Coordinates extracted from %s', self.template_file)

        try:
            with h5py.File(self.h5_file, mode='w') as dst:
                profile = deepcopy(self.profile)
                profile.pop("dtype", None)
                dst.attrs['profile'] = json.dumps(profile)
                logger.debug('\t- Default profile:\n%s',
                             json.dumps(profile, indent=4))
                dst.attrs['shape'] = self.shape
                logger.debug('\t- Default shape:\n%s', self.shape)
                dst.attrs['chunks'] = self._chunks
                logger.debug('\t- Default chunks:\n%s', self._chunks)

                profile['dtype'] = str(np.float32)
                profile['count'] = 1
                ds = dst.create_dataset(self.LATITUDE, shape=lat.shape,
                                        dtype=np.float32, data=lat,
                                        chunks=self._chunks)
                ds.attrs['profile'] = json.dumps(profile)
                logger.debug('\t- latitude coordinates created')

                ds = dst.create_dataset(self.LONGITUDE, shape=lon.shape,
                                        dtype=np.float32, data=lon,
                                        chunks=self._chunks)
                ds.attrs['profile'] = json.dumps(profile)
                logger.debug('\t- longitude coordinates created')
        except Exception as e:
            logger.error("Error initializing %s", self.h5_file)
            logger.exception(e)
            if os.path.exists(self.h5_file):
                os.remove(self.h5_file)

    def write_layer_to_h5(self, values, layer_name, profile=None,
                          description=None, scale_factor=None):
        """Write a layer to the HDF5 file.

        Parameters
        ----------
        values : ndarray
            Layer data.
        layer_name : str
            Dataset name in HDF5 file.
        profile : dict, optional
            Layer profile (attributes). If ``None``, the profile from
            the Layered HDF5 file is used instead.
        description : str, optional
            Description of layer being added. By default, ``None``,
            which does not store a description in the layer attributes.
        scale_factor : int | float, optional
            Scale factor to use to scale geotiff data when added to the
            HDF5 file. By default, ``None``, which does not scale the
            values.
        """
        if not Path(self.h5_file).exists():
            self.create_new(overwrite=False)

        if values.ndim < 3:
            values = np.expand_dims(values, 0)

        if values.shape[1:] != self.shape:
            raise ValueError(f'Shape of provided data {values.shape[1:]} does '
                             f'not match template raster {self.shape}.')

        chunks = self._chunks
        if len(chunks) < 3:
            chunks = (1, ) + chunks

        with h5py.File(self.h5_file, mode='a') as f:
            if layer_name in f:
                ds = f[layer_name]
                ds[...] = values
                logger.debug('\t- %s values replaced', layer_name)
            else:
                ds = f.create_dataset(layer_name, shape=values.shape,
                                      dtype=values.dtype, chunks=chunks,
                                      data=values)
                logger.debug('\t- %s created and loaded', layer_name)

            profile = deepcopy(profile or self.profile)
            profile['dtype'] = str(values.dtype)
            profile['count'] = 1
            ds.attrs['profile'] = json.dumps(profile)
            logger.debug('\t- Unique profile for %s added:\n%s',
                         layer_name, json.dumps(profile, indent=4))
            if description is not None:
                ds.attrs['description'] = description
                logger.debug('\t- Description for %s added:\n%s',
                             layer_name, description)

            if scale_factor is not None:
                ds.attrs['scale_factor'] = scale_factor
                logger.debug('\t- scale_factor for %s added:\n%.2f',
                             layer_name, scale_factor)

    def write_geotiff_to_h5(self, geotiff, layer_name, check_tiff=True,
                            transform_atol=0.01, description=None,
                            scale_factor=None, dtype='int16', replace=True):
        """Transfer GeoTIFF to HDF5 confirming it matches existing layers.

        Parameters
        ----------
        geotiff : str
            Path to GeoTIFF file.
        layer_name : str
            Name of layer to be written to HDF5 file.
        check_tiff : bool, optional
            Option to check GeoTIFF profile, CRS, and shape against
            layered HDF5 profile, CRS, and shape. By default, ``True``.
        transform_atol : float, optional
            Absolute tolerance parameter when comparing GeoTIFF
            transform data. By default, ``0.01``.
        description : str, optional
            Optional description of layer. By default, ``None``.
        scale_factor : int | float, optional
            Scale factor to use to scale GeoTIFF data when added to the
            layered HDF5 file, by default None
        dtype : str, optional
            Dtype to save GeoTIFF data as in the layered HDF5 file. Only
            used when `scale_factor` input is not ``None``.
            By default, ``"int16"``.
        replace : bool, optional
            Option to replace existing layer (if any).
            By default, ``True``.
        """
        if not Path(self.h5_file).exists():
            if self.template_file == self.h5_file:
                self.template_file = geotiff
            self.create_new(overwrite=False)

        self._warn_or_error_for_existing_layer(layer_name, replace)

        logger.debug('\t- %s being extracted from %s and added to %s',
                     layer_name, geotiff, self.h5_file)

        if check_tiff:
            check_geotiff(self, geotiff, chunks=self._chunks,
                          transform_atol=transform_atol)

        with Geotiff(geotiff, chunks=self._chunks) as tif:
            profile, values = tif.profile, tif.values

        if scale_factor is not None:
            attrs = {'scale_factor': scale_factor}
            values = Outputs._check_data_dtype(layer_name, values, dtype,
                                               attrs=attrs)

        self.write_layer_to_h5(values, layer_name, profile=profile,
                               description=description,
                               scale_factor=scale_factor)

    def _warn_or_error_for_existing_layer(self, layer_name, replace):
        """Warn about existing layers. """
        if layer_name not in self.layers:
            return

        msg = "{} is already present in {}".format(layer_name, self.h5_file)
        if replace:
            msg += " and will be replaced"
            logger.warning(msg)
            warn(msg)
        else:
            msg += ", to 'replace' set to True"
            logger.error(msg)
            raise KeyError(msg)

    def layer_to_geotiff(self, layer, geotiff):
        """Extract layer from HDF5 file and write to GeoTIFF file.

        Parameters
        ----------
        layer : str
            Layer to extract,
        geotiff : str
            Path to output GeoTIFF file.
        """
        profile, values = self[layer]
        logger.debug('\t- Writing %s to %s', layer, geotiff)
        Geotiff.write(geotiff, profile, values)

    def save_data_using_h5_profile(self, data, geotiff):
        """Write to GeoTIFF file.

        Parameters
        ----------
        layer : str
            Layer to extract,
        geotiff : str
            Path to output GeoTIFF file.
        """
        dtype = data.dtype.name
        if dtype == 'bool':
            dtype = 'uint8'
        profile = deepcopy(self.profile)

        Geotiff.write(geotiff, profile, data, dtype=dtype)

    def load_data_using_h5_profile(self, geotiff, band=1, reproject=False,
                                   skip_profile_test=False):
        """Load GeoTIFF data, converting to H5 profile if necessary.

        Parameters
        ----------
        geotiff : str
            Path to GeoTIFF from which data should be read.
        band : int, optional
            Band to load from GeoTIFF. By default, ``1``.
        reproject : bool, optional
            Reproject raster to standard CRS and transform if True.
            By default, ``False``.
        skip_profile_test: bool, optional
            Skip checking that shape, transform, and CRS match template raster
            if ``True``. By default, ``False``.

        Returns
        -------
        array-like
            Raster data.
        """

        with Geotiff(geotiff) as geo:
            data = geo.values[band - 1]
            src_profile = geo.profile

        if skip_profile_test:
            return data

        try:
            check_geotiff(self, geotiff, chunks=self._chunks)
        except ProfileCheckError as err:
            if reproject:
                logger.debug('Profile of %s does not match template, '
                             'reprojecting', geotiff)
                data = self.reproject(data, src_profile, dtype=data.dtype,
                                      init_dest=0)
            else:
                raise err

        return data

    def reproject(self, src_raster, src_profile, dtype='float32',
                  init_dest=-1.0):
        """Reproject a raster onto the template raster and transform.

        Parameters
        ----------
        src_raster : array-like
            Source raster.
        src_profile : dict
            Source raster profile.
        dtype : np.dtype, optional
            Data type for destination raster. By default, ``"float32"``.
        init_des : float, optional
            Value for cells outside of boundary of src_raster.
            By default, ``-1.0``.

        Returns
        -------
        array-like
            Source data reprojected into the template projection.
        """
        dest_raster = np.zeros(self.shape, dtype=dtype)
        reproject(src_raster,
                  destination=dest_raster,
                  src_transform=src_profile['transform'],
                  src_crs=src_profile['crs'],
                  dst_transform=self.profile['transform'],
                  dst_crs=self.profile['crs'],
                  num_threads=4,
                  resampling=Resampling.nearest,
                  INIT_DEST=init_dest)
        return dest_raster

    def layers_to_h5(self, layers, replace=True, check_tiff=True,
                     transform_atol=0.01, descriptions=None,
                     scale_factors=None):
        """Transfer GeoTIFF layers into layered HDF5 file.

        If layered HDF5 file does not exist, it is created and
        populated.

        Parameters
        ----------
        layers : list | dict
            List of GeoTIFFs to load or dictionary mapping GeoTIFFs to
            the layers to load.
        replace : bool, optional
            Option to replace existing layers if needed.
            By default, ``True``
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against layered
            HDF5 profile and coordinates. By default, ``True``.
        transform_atol : float, optional
            Absolute tolerance parameter when comparing GeoTIFF and
            layered HDF5 transforms. By default, ``0.01``.
        description : dict, optional
            Mapping of layer name to layer description of layers.
            By default, ``None``, which does not store any descriptions.
        scale_factor : dict, optional
            Scale factors and dtypes to use when scaling given layers.
            By default, ``None``, which does not apply any scale
            factors.
        """
        if isinstance(layers, list):
            layers = {os.path.basename(lyr).split('.')[0]: lyr
                      for lyr in layers}

        if descriptions is None:
            descriptions = {}

        if scale_factors is None:
            scale_factors = {}

        logger.info('Moving layers to %s', self.h5_file)
        for layer_name, geotiff in layers.items():
            logger.info('- Transferring %s', layer_name)
            description = descriptions.get(layer_name, None)
            scale = scale_factors.get(layer_name, None)
            if scale is not None:
                scale_factor = scale['scale_factor']
                dtype = scale['dtype']
            else:
                scale_factor = None
                dtype = None

            self.write_geotiff_to_h5(geotiff, layer_name,
                                     check_tiff=check_tiff,
                                     transform_atol=transform_atol,
                                     description=description,
                                     scale_factor=scale_factor,
                                     dtype=dtype, replace=replace)

    def extract_layers(self, layers):
        """Extract layers from HDF5 file and save to disk as GeoTIFFs.

        Parameters
        ----------
        layers : dict
            Dictionary mapping layer names to GeoTIFF files to create.
        """
        logger.info('Extracting layers from %s', self.h5_file)
        for layer_name, geotiff in layers.items():
            logger.info('- Extracting %s', layer_name)
            self.layer_to_geotiff(layer_name, geotiff)

    def extract_all_layers(self, out_dir, extract_lat_lon=False):
        """Extract all layers from HDF5 file and save to disk as GeoTIFFs.

        Parameters
        ----------
        out_dir : str
            Path to output directory into which layers should be saved
            as GeoTIFFs.
        extract_lat_lon : bool, default=False
            Option to extract latitude and longitude layers in addition
            to all other layers. By default, ``False``.
        """
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        layer_names = self.layers
        if not extract_lat_lon:
            layer_names = [layer for layer in layer_names
                           if layer not in {self.LATITUDE, self.LONGITUDE}]

        layers = {layer_name: os.path.join(out_dir, f"{layer_name}.tif")
                  for layer_name in layer_names}
        self.extract_layers(layers)


class LayeredTransmissionH5(LayeredH5):
    """
    Handle reading and writing H5 files and GeoTiffs
    """

    def __init__(self, h5_file=None, hsds=False, chunks=(128, 128),
                 template_file=None, layer_dir='.', block_size=None):
        """

        Parameters
        ----------
        h5_file : path-like, optional
            Path to layered transmission HDF5 file. If this file is to
            be created, a `template_file` must be provided (and must
            exist on disk). Otherwise, the `template_file` input can be
            ignored and this input will be used as the template file.
            This input can be set to `None` if only the tiff conversion
            utilities are required, but the `template_file` input must
            be provided in this case. By default, ``None``.
        hsds : bool, optional
            Boolean flag to use h5pyd to handle HDF5 'files' hosted on
            AWS behind HSDS. By default, ``False``.
        chunks : tuple, optional
            Chunk size of exclusions in HDF5 file and any output
            GeoTIFFs. By default, ``(128, 128)``.
        template_file : path-like, optional
            Path to template GeoTIFF (``*.tif`` or ``*.tiff``) or HDF5
            (``*.h5``) file containing the profile and transform to be
            used for the layered transmission file. If ``None``, then
            the `h5_file` input is used as the template. If ``None`` and
            the `h5_file` input is also ``None``, an error is thrown.
            By default, ``None``.
        layer_dir : path-like, optional
            Directory to search for layers in, if not found in current
            directory. By default, ``'.'``.
        block_size : int, optional
            Optional block size to use when building lat/lon datasets.
            Setting this value can help reduce memory issues when
            building a ``LayeredH5`` file. If ``None``, the lat/lon
            arrays are processed in full. By default, ``None``.
        """
        super().__init__(h5_file=h5_file, hsds=hsds, chunks=chunks,
                         template_file=template_file, block_size=block_size)
        self._layer_dir = layer_dir
        if self.h5_file is None and self.template_file is None:
            msg = "One of `h5_file` or `template_file` must be provided!"
            logger.error(msg)
            raise ValueError(msg)

    def load_data_using_h5_profile(self, geotiff, band=1, reproject=False,
                                   skip_profile_test=False):
        """Load GeoTIFF data, converting to H5 profile if necessary.

        Parameters
        ----------
        geotiff : str
            Path to GeoTIFF from which data should be read. If just the
            file name is provided, the class `layer_dir` attribute value
            is prepended to get the full path.
        band : int, optional
            Band to load from GeoTIFF. By default, ``1``.
        reproject : bool, optional
            Reproject raster to standard CRS and transform if True.
            By default, ``False``.
        skip_profile_test: bool, optional
            Skip checking that shape, transform, and CRS match template raster
            if ``True``. By default, ``False``.

        Returns
        -------
        array-like
            Raster data.
        """
        full_fname = geotiff
        if not Path(full_fname).exists():
            full_fname = os.path.join(self._layer_dir, geotiff)
            if not Path(full_fname).exists():
                raise FileNotFoundError(f'Unable to find file {geotiff}')

        skip_test = skip_profile_test
        return super().load_data_using_h5_profile(geotiff=full_fname,
                                                  band=band,
                                                  reproject=reproject,
                                                  skip_profile_test=skip_test)


def check_geotiff(h5, geotiff, chunks=(128, 128), transform_atol=0.01):
    """Compare GeoTIFF with exclusion layer and raise errors if mismatch.

    Parameters
    ----------
    h5 : :class:`LayeredH5`
        ``LayeredH5`` instance containing `shape`, `profile`, and
        `template_file` attributes.
    geotiff : str
        Path to GeoTIFF file.
    chunks : tuple
        Chunk size of exclusions in GeoTIFF,
    transform_atol : float
        Absolute tolerance parameter when comparing GeoTIFF transform
        data.

    Returns
    -------
    profile : dict
        GeoTIFF profile (attributes).
    values : ndarray
        GeoTIFF data.

    Raises
    ------
    ProfileCheckError
        If shape, profile, or transform don;t match between HDF5 and
        GeoTIFF file.
    """
    with Geotiff(geotiff, chunks=chunks) as tif:
        if tif.bands > 1:
            msg = "{} contains more than one band!".format(geotiff)
            logger.error(msg)
            raise ProfileCheckError(msg)

        if not np.array_equal(h5.shape, tif.shape):
            msg = ('Shape of exclusion data in {} and {} do not '
                   'match!'.format(geotiff, h5.template_file))
            logger.error(msg)
            raise ProfileCheckError(msg)

        h5_crs = CRS.from_string(h5.profile['crs']).to_dict()
        tif_crs = CRS.from_string(tif.profile['crs']).to_dict()
        if not crs_match(h5_crs, tif_crs):
            msg = ('Geospatial "CRS" in {} and {} do not match!\n {} !=\n {}'
                   .format(geotiff, h5.template_file, tif_crs, h5_crs))
            logger.error(msg)
            raise ProfileCheckError(msg)

        if not np.allclose(h5.profile['transform'], tif.profile['transform'],
                           atol=transform_atol):
            msg = ('Geospatial "transform" in {} and {} do not match!'
                   '\n {} !=\n {}'
                   .format(geotiff, h5.template_file, h5.profile['transform'],
                           tif.profile['transform']))
            logger.error(msg)
            raise ProfileCheckError(msg)


def crs_match(baseline_crs, test_crs, ignore_keys=('no_defs',)):
    """Compare baseline and test CRS values.

    Parameters
    ----------
    baseline_crs : dict
        Baseline CRS to use a truth, must be a dict
    test_crs : dict
        Test CRS to compare with baseline, must be a dictionary.
    ignore_keys : tuple, optional
        Keys to not check. By default, ``('no_defs',)``.

    Returns
    -------
    crs_match : bool
        ``True`` if crs' match, ``False`` otherwise
    """
    for k, true_v in baseline_crs.items():
        if k not in ignore_keys:
            test_v = test_crs.get(k, true_v)
            if true_v != test_v:
                return False

    return True



# -*- coding: utf-8 -*-
"""
reVX command line interface (CLI).
"""
import click
import logging
import os
import json
from pathlib import Path

from gaps.cli import CLICommandFromFunction, as_click_command

from rex.utilities.cli_dtypes import STR, STRLIST, FLOAT
from rex.utilities.loggers import init_logger
from rex.utilities.utilities import safe_json_load
from reV.supply_curve.exclusions import ExclusionMaskFromDict, ExclusionLayers

from reVX.offshore.dist_to_ports_converter import DistToPortsConverter
from reVX.handlers.geotiff import Geotiff
from reVX.handlers.layered_h5 import LayeredH5
from reVX.utilities.forecasts import FcstUtils
from reVX.utilities.output_extractor import output_extractor
from reVX.utilities.region_classifier import RegionClassifier
from reVX.utilities.reeds_cols import add_reeds_columns
from reVX.setbacks.setbacks_converter import SetbacksConverter

from reVX import __version__

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, verbose):
    """reVX command line interface."""
    ctx.ensure_object(dict)
    if verbose:
        log_level = 'DEBUG'
    else:
        log_level = 'INFO'

    init_logger('reVX', log_level=log_level)


@main.command()
@click.option('--meta_path', '-mp', required=True,
              prompt='Path to meta CSV file, resource .h5 file',
              type=click.Path(exists=True),
              help=('Path to meta CSV file, resource .h5 file containing '
                    'lat/lon points'))
@click.option('--regions_path', '-rp', required=True,
              prompt='Path to regions shapefile',
              type=click.Path(exists=True),
              help=('Path to regions shapefile containing labeled geometries'))
@click.option('--regions_label', '-rl', default=None, type=STR,
              show_default=True,
              help=('Attribute to use as label in the regions shapefile'))
@click.option('--fout', '-o', required=True,
              prompt='Output CSV file path',
              type=click.Path(),
              help=('Output CSV file path for labeled meta CSV file'))
@click.option('--force', '-f', is_flag=True,
              help='Force outlier classification by finding nearest.')
def region_classifier(meta_path, regions_path, regions_label, fout,
                      force):
    """
    Region Classifier
    - Used to classify meta points with a label from a shapefile
    """
    RegionClassifier.run(meta_path=meta_path,
                         regions=regions_path,
                         regions_label=regions_label,
                         force=force, fout=fout)


@main.command()
@click.option('--my_fpath', '-src', required=True,
              type=click.Path(exists=True),
              help='Path to multi-year output .h5 file')
@click.option('--out_fpath', '-out', required=True,
              type=click.Path(),
              help='Path to output .h5 file')
@click.option('--year', '-yr', default=None, type=STR,
              show_default=True,
              help='Year to extract, if None parse from out_fpath')
def extract_output_year(my_fpath, out_fpath, year):
    """
    Extract all datasets for a give year from multi-year output file
    """
    output_extractor(my_fpath, out_fpath, year=year)


@main.command()
@click.option('--fcst_h5', '-fcst', required=True,
              type=click.Path(exists=True),
              help="Path to forecast .h5 file")
@click.option('--fcst_dset', '-fdset', required=True, type=str,
              help="Dataset to correct")
@click.option('--out_h5', '-out', required=True, type=click.Path(),
              help="Output path for corrected .h5 file")
@click.option('--actuals_h5', '-actuals', type=click.Path(exists=False),
              default=None, show_default=True,
              help="Path to forecast to .h5 file, by default None")
@click.option('--actuals_dset', '-adset', default=None, type=STR,
              show_default=True,
              help="Actuals dataset, by default None")
@click.option('--fcst_perc', '-perc', default=None, type=FLOAT,
              show_default=True,
              help=("Percentage of forecast to use for blending, by default "
                    "None"))
def correct_forecast(fcst_h5, fcst_dset, out_h5, actuals_h5, actuals_dset,
                     fcst_perc):
    """
    Bias correct and blend (if requested) forecasts using actuals:
    - Bias correct forecast data using bias correction factor:
    total actual generation / total forecasted generation
    - Blend fcst_perc of forecast generation with (1 - fcst_perc) of
    actuals generation
    """
    FcstUtils.correct(fcst_h5, fcst_dset, out_h5, actuals_h5=actuals_h5,
                      actuals_dset=actuals_dset, fcst_perc=fcst_perc)


@main.group()
@click.option('--excl_h5', '-h5', required=True, type=click.Path(exists=False),
              help=("Path to .h5 file containing or to contain exclusion "
                    "layers"))
@click.pass_context
def exclusions(ctx, excl_h5):
    """
    Extract from or create exclusions .h5 file
    """
    ctx.obj['EXCL_H5'] = excl_h5


@exclusions.command()
@click.option('--layers', '-l', required=True, type=click.Path(exists=True),
              help=(".json file containing mapping of layer names to geotiffs."
                    " Json can also contain layer descriptions and/or "
                    "scale factors"))
@click.option('-check_tiff', '-ct', is_flag=True,
              help=("Flag to check tiff profile, CRS, and shape against "
                    "exclusion .h5 profile, CRS, and shape"))
@click.option('--setbacks', '-sb', is_flag=True,
              help=("Flag to convert setbacks to exclusion layers"))
@click.option('--distance_to_ports', '-dtp', is_flag=True,
              help=("Flag to convert distances to ports to exclusion layers"))
@click.option('--transform_atol', '-tatol', default=0.01, type=float,
              show_default=True,
              help=("Absolute tolerance parameter when comparing geotiff "
                    "transform data."))
@click.option('--purge', '-r', is_flag=True,
              help="Remove existing .h5 file before loading layers")
@click.pass_context
def layers_to_h5(ctx, layers, check_tiff, setbacks, distance_to_ports,
                 transform_atol, purge):
    """
    Add layers to exclusions .h5 file
    """
    excl_h5 = ctx.obj['EXCL_H5']
    if purge and os.path.isfile(excl_h5):
        os.remove(excl_h5)

    inputs = safe_json_load(layers)
    layers = inputs['layers']
    descriptions = inputs.get('descriptions')
    scale_factors = inputs.get('scale_factors')

    if setbacks and distance_to_ports:
        msg = ('Both the setbacks "--setbacks/-sb" and distance to ports '
               '"--distance_to_ports/-dtp" flags were supplied! Setbacks and '
               'distance to ports and computed differently, please only '
               'select one!')
        logger.error(msg)
        raise RuntimeError(msg)

    if setbacks:
        are_inclusion_layers = inputs.get('are_inclusion_layers', False)
        converter = SetbacksConverter(excl_h5)
        converter.layers_to_h5(layers, check_tiff=check_tiff,
                               are_inclusion_layers=are_inclusion_layers,
                               transform_atol=transform_atol,
                               descriptions=descriptions,
                               scale_factors=scale_factors)
    elif distance_to_ports:
        converter = DistToPortsConverter(excl_h5)
        converter.layers_to_h5(layers, check_tiff=check_tiff,
                               transform_atol=transform_atol,
                               descriptions=descriptions,
                               scale_factors=scale_factors)
    else:
        converter = LayeredH5(excl_h5)
        converter.layers_to_h5(layers, check_tiff=check_tiff,
                               transform_atol=transform_atol,
                               descriptions=descriptions,
                               scale_factors=scale_factors)


@exclusions.command()
@click.option('--out_dir', '-o', required=True, type=click.Path(exists=True),
              help=("Output directory to save layers into"))
@click.option('--layers', '-l', default=None, type=STRLIST,
              show_default=True,
              help=("List of layers to extract, if None extract all"))
@click.option('--hsds', '-hsds', is_flag=True,
              help="Extract layers from HSDS")
@click.pass_context
def layers_from_h5(ctx, out_dir, layers, hsds):
    """
    Extract layers from excl .h5 file and save to disk as geotiffs
    """
    excl_h5 = ctx.obj['EXCL_H5']
    if layers is not None:
        layers = {layer: os.path.join(out_dir, "{}.tif".format(layer))
                  for layer in layers}
        LayeredH5(excl_h5, hsds=hsds).extract_layers(layers)
    else:
        LayeredH5(excl_h5, hsds=hsds).extract_all_layers(out_dir)


@exclusions.command()
@click.option('--excl_dict_fpath', '-ed', required=True,
              type=click.Path(exists=True),
              help=('Path to JSON file containing the ``"excl_dict"`` '
                    'key which points to the exclusion dictionary defining '
                    'the mask that should be generated. A typical reV '
                    'aggregation config satisfied this requirement. If this '
                    'file also contains an ``"excl_fpath"`` key, the value '
                    'from the file will override the ``--excl_h5`` CLI '
                    'argument input.'))
@click.option('--out', '-o', required=True, type=STR,
              help=('Output name. If this string value ends in ".tif" '
                    'or ".tiff", this input is assumed to be a path to an '
                    'output tiff file, and the mask will be written to that '
                    'destination. Otherwise, this input is assumed to be the '
                    'name of the layer in the exclusion file to write the '
                    'mask to.'))
@click.option('--min_area', '-ma', default=None, type=FLOAT,
              help=('Minimum required contiguous area in sq-km.'))
@click.option('--kernel', '-k', type=STR, default='queen',
              show_default=True,
              help=('Contiguous filter method to use on final exclusion.'))
@click.option('--hsds', '-hsds', is_flag=True,
              help=('Flag to use h5pyd to handle .h5 domain hosted on AWS '
                    'behind HSDS'))
@click.pass_context
def mask(ctx, excl_dict_fpath, out, min_area, kernel, hsds):
    """
    Compute Setbacks locally
    """
    log_level = "DEBUG" if ctx.obj.get('VERBOSE') else "INFO"
    init_logger('reV', log_level=log_level)
    init_logger('reVX', log_level=log_level)

    logger.info("Calculating exclusion mask from {!r}".format(excl_dict_fpath))

    with open(excl_dict_fpath, 'r') as fh:
        config = json.load(fh)

    excl_fpath = config.get("excl_fpath", ctx.obj['EXCL_H5'])
    logger.debug("Exclusion filepath(s): {!r}".format(excl_fpath))

    excl_dict = config['excl_dict']
    logger.debug("Exclusion dictionary: {!r}".format(excl_dict))
    mask_ = ExclusionMaskFromDict.run(excl_fpath, layers_dict=excl_dict,
                                      min_area=min_area, kernel=kernel,
                                      hsds=hsds)

    with ExclusionLayers(excl_fpath, hsds=hsds) as f:
        profile = f.profile

    if out.endswith(".tif") or out.endswith(".tiff"):
        out = Path(out).resolve().as_posix()
        logger.info("Writing mask to {!r}".format(out))
        Geotiff.write(out, profile, mask_)
    else:
        if isinstance(excl_fpath, list):
            excl_fpath = excl_fpath[0]
        logger.info("Writing mask to layer {!r} in {!r}"
                    .format(out, excl_fpath))
        desc = ("Exclusion mask computed from exclusion dictionary: {!r}"
                .format(excl_dict))
        LayeredH5(excl_fpath).write_layer_to_h5(mask_, out, profile,
                                                description=desc)


def _reeds_cols_preprocessor(config):
    """Ensure supply_curve_fpath and out_fpath inputs are lists """
    for key in ["supply_curve_fpath", "out_fp"]:
        if isinstance(config.get(key), str):
            config[key] = [config[key]]

    if config.get("out_fp") is None:
        config["out_fp"] = config["supply_curve_fpath"]

    if len(config["supply_curve_fpath"]) != len(config["out_fp"]):
        raise ValueError("Length of 'out_fp' must match length of "
                         "'supply_curve_fpath'")

    return config


command = CLICommandFromFunction(add_reeds_columns,
                                 name="add-reeds-cols",
                                 split_keys=[("supply_curve_fpath", "out_fp")],
                                 config_preprocessor=_reeds_cols_preprocessor)
main.add_command(as_click_command(command))


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reVX CLI')
        raise



# -*- coding: utf-8 -*-
"""
Setback exclusion computation classes
"""
import logging

from reVX.setbacks.base import AbstractBaseSetbacks, BUFFERS, FEATURE_FILTERS


logger = logging.getLogger(__name__)


def _format_str(in_str):
    """Format a string like regulation class does. """
    return in_str.strip().lower().replace("-", " ").replace("_", " ")


def _camel_case_str(in_str):
    """Format a string into CamelCase"""
    return _format_str(in_str).title().replace(" ", "")


def setbacks_calculator(feature_type, buffer_type="default",
                        feature_filter_type="centroid",
                        feature_subtypes_to_exclude=None,
                        num_features_per_worker=10_000):
    """Generate a setback calculator for a particular feature.

    Parameters
    ----------
    feature_type : str | list
        String or list of strings that represent how this feature is
        distinguished in the local regulations file. In particular,
        this string (or list of strings) should match the values under
        the ``Feature Type`` column of the setback regulations file
        corresponding to this feature. For example, specifying
        ``feature_type="property line"`` would mean that this setback
        calculator looks for ``"property line"`` regulations (i.e.
        parcel setbacks) in the input regulations file. This input is
        case insensitive (i.e. "ROADS", "Roads", "roads", and "rOAds"
        are all equivalent), and spaces, dashes, and underscores are
        interchangeable (i.e. "property line", "property_line", and
        "property-line" are all equivalent).
    buffer_type : str, optional
        Name of the buffer type to use for this setback calculation. See
        :attr:`~reVX.setbacks.base.BUFFERS` for all available options -
        this input must be a key from that dictionary. In most cases,
        the default positive buffer is suitable.
        By default, ``"default"``.
    feature_filter_type : str, optional
        Name of the feature filter type to use for this setback
        calculation. Typically, features like structures or parcels are
        filtered using "centroid", which only applies setback
        regulations to features whose centroid is within the county with
        the ordinance. On the other hand, features such as roads,
        railroads transmission lines, etc., are filtered using "clip",
        which clips the feature to the county boundaries before applying
        a setback. This value must be a key in the
        :attr:`~reVX.setbacks.base.FEATURE_FILTERS` dictionary.
        By default, ``"centroid"``.
    feature_subtypes_to_exclude : list | set, optional
        A list of strings specifying the feature subtypes that should be
        excluded from the local regulations calculation.
        By default, ``None``.
    num_features_per_worker : int, optional
        Number of features each worker (CPU processor) on a node
        processes. By default, ``10,000``.

    Returns
    -------
    class
        A class object that can be used to instantiate the setbacks
        calculator. The initializer parameters are identical to that of
        :class:`~reVX.setbacks.base.AbstractBaseSetbacks`.
    """
    if isinstance(feature_type, str):
        feature_type = [feature_type]

    camel_case_feature = "".join(map(_camel_case_str, sorted(feature_type)))
    feature_type = set(map(_format_str, feature_type))

    if buffer_type not in BUFFERS:
        msg = ("Unknown buffer type specified: {!r}. Must be one of {}"
               .format(buffer_type, list(BUFFERS)))
        logger.error(msg)
        raise ValueError(msg)

    if feature_filter_type not in FEATURE_FILTERS:
        msg = ("Unknown feature filter type specified: {!r}. Must be one of {}"
               .format(feature_filter_type, list(FEATURE_FILTERS)))
        logger.error(msg)
        raise ValueError(msg)

    feature_subtypes_to_exclude = set(feature_subtypes_to_exclude or set())
    num_features_per_worker = int(max(0, num_features_per_worker) or 10_000)
    class_name = "{}Setbacks".format(camel_case_feature)
    class_attrs = {"FEATURE_TYPES": feature_type,
                   "FEATURE_SUBTYPES_TO_EXCLUDE": feature_subtypes_to_exclude,
                   "BUFFER_TYPE": buffer_type,
                   "FEATURE_FILTER_TYPE": feature_filter_type,
                   "NUM_FEATURES_PER_WORKER": num_features_per_worker}

    return type(class_name, (AbstractBaseSetbacks,), class_attrs)


SETBACK_SPECS = {
    "parcel": {
        "feature_type": "property line",
        "buffer_type": "parcel",
        "feature_filter_type": "centroid",
        "feature_subtypes_to_exclude": None,
        "num_features_per_worker": 10_000,
    },
    "rail": {
        "feature_type": "railroads",
        "buffer_type": "default",
        "feature_filter_type": "clip",
        "feature_subtypes_to_exclude": None,
        "num_features_per_worker": 10_000,
    },
    "road": {
        "feature_type": ['roads', 'highways', 'highways 111'],
        "buffer_type": "default",
        "feature_filter_type": "clip",
        "feature_subtypes_to_exclude": None,
        "num_features_per_worker": 10_000,
    },
    "structure": {
        "feature_type": "structures",
        "buffer_type": "default",
        "feature_filter_type": "centroid",
        "feature_subtypes_to_exclude": ["Occupied Community Buildings"],
        "num_features_per_worker": 10_000,
    },
    "transmission": {
        "feature_type": "transmission",
        "buffer_type": "default",
        "feature_filter_type": "clip",
        "feature_subtypes_to_exclude": None,
        "num_features_per_worker": 10_000,
    },
    "water": {
        "feature_type": "water",
        "buffer_type": "default",
        "feature_filter_type": "clip",
        "feature_subtypes_to_exclude": None,
        "num_features_per_worker": 700,
    },
}
"""Default Setback Calculator Specifications.

These calculators are available by default. To use them, simply match
the key in this dictionary with your dataset path.
"""


SETBACKS = {key: setbacks_calculator(**kwargs)
            for key, kwargs in SETBACK_SPECS.items()}



# -*- coding: utf-8 -*-
"""
Compute setbacks exclusions
"""
from warnings import warn
import logging

from reVX.utilities.regulations import AbstractBaseRegulations


logger = logging.getLogger(__name__)


class SetbackRegulations(AbstractBaseRegulations):
    """Setback regulation values. """

    def __init__(self, base_setback_dist, regulations_fpath=None,
                 multiplier=None):
        """
        Parameters
        ----------
        base_setback_dist : float | int
            Base setback distance (m). This value will be used to
            calculate the setback distance when a multiplier is provided
            either via the ``regulations_fpath`` csv or the
            ``multiplier`` input. In these cases, the setbacks will be
            calculated using ``base_setback_dist * multiplier``.
        regulations_fpath : str | None, optional
            Path to regulations ``.csv`` or ``.gpkg`` file. At a
            minimum, this file must contain the following columns:

                - ``Feature Type``: Contains labels for the type of
                  setback that each row represents. This should be a
                  `"feature_type"` label that can be found in the
                  :attr:`~reVX.setbacks.setbacks.SETBACK_SPECS`
                  dictionary (e.g. ``"structures"``, ``"roads"``,
                  ``"water"``, etc.), unless you have created your own
                  setback calculator using
                  :func:`~reVX.setbacks.setbacks.setbacks_calculator`,
                  in which case this label can match the `feature_type`
                  input you used for that function call.
                - ``Feature Subtype``: Contains labels for feature
                  subtypes. The feature subtypes are only used for
                  down-selecting the local regulations that should be
                  applied for a particular feature, so often you can
                  leave this blank or set it to ``None``. If you do
                  specify this value, it should be a
                  `"feature_subtypes_to_exclude"` label that can be
                  found in the
                  :attr:`~reVX.setbacks.setbacks.SETBACK_SPECS`
                  dictionary,  unless you have created your own setback
                  calculator using
                  :func:`~reVX.setbacks.setbacks.setbacks_calculator`,
                  in which case this label can match the
                  `feature_subtypes_to_exclude` input you used for that
                  function call.
                - ``Value Type``: Specifies wether the value is a
                  multiplier or static height. See below for more info.
                - ``Value``: Numeric value of the setback or multiplier.
                - ``FIPS``: Specifies a unique 5-digit code for each
                  county (this can be an integer - no leading zeros
                  required). This is used to match the county
                  regulations to the county's spatial extent.

            Valid options for the ``Value Type`` are (case-insensitive;
            dashes, underscores, and spaces are interchangeable):

                - "Structure Height Multiplier"
                - "Meters"

            If this input is ``None``, a generic setback of
            ``base_setback_dist * multiplier`` is used.
            By default ``None``.
        multiplier : int | float | str | None, optional
            A setback multiplier to use if regulations are not supplied.
            This multiplier will be applied to the ``base_setback_dist``
            to calculate the setback. If supplied along with
            ``regulations_fpath``, this input will be used to apply a
            setback to all counties not listed in the regulations file.
            By default ``None``.
        """
        self._multi = multiplier
        super().__init__(generic_regulation_value=base_setback_dist,
                         regulations_fpath=regulations_fpath)

    def _preflight_check(self, regulations_fpath):
        """Apply preflight checks to the regulations path and multiplier.

        Run preflight checks on setback inputs:
        1) Ensure either a regulations .csv or
           a setback multiplier (or both) is provided
        2) Ensure regulations has county FIPS, map regulations to county
           geometries from exclusions .h5 file

        Parameters
        ----------
        regulations_fpath : str | None
            Path to regulations .csv file, if `None`, create global
            setbacks.
        """
        super()._preflight_check(regulations_fpath)

        if self._multi is not None:
            logger.debug('Computing setbacks using base setback distance '
                         'multiplier of {}'.format(self._multi))

        if not regulations_fpath and self._multi is None:
            msg = ('Computing setbacks requires a regulations '
                   '.csv file and/or a generic multiplier!')
            logger.error(msg)
            raise RuntimeError(msg)

    @property
    def base_setback_dist(self):
        """float: The base setback distance, in meters. """
        return self._generic_regulation_value

    @property
    def multiplier(self):
        """int | float: Generic setback multiplier. """
        return self._multi

    @property
    def generic(self):
        """float | None: Regulation value used for global regulations. """
        if self.multiplier is None:
            setback = None
        else:
            setback = self.base_setback_dist * self.multiplier

        return setback

    def _county_regulation_value(self, county_regulations):
        """Retrieve county regulation setback. """
        setback_type = county_regulations["Value Type"]
        setback = float(county_regulations["Value"])
        if setback_type == "structure height multiplier":
            setback *= self.base_setback_dist
        elif setback_type != "meters":
            msg = ("Cannot create setback for {}, expecting "
                   '"Structure Height Multiplier", or '
                   '"Meters", but got {!r}'
                   .format(county_regulations["County"], setback_type))
            logger.warning(msg)
            warn(msg)
            return
        return setback


class WindSetbackRegulations(SetbackRegulations):
    """Wind setback regulation setback values. """

    MULTIPLIERS = {'high': 3, 'moderate': 1.1}
    """Named generic multipliers. """

    def __init__(self, hub_height, rotor_diameter, regulations_fpath=None,
                 multiplier=None):
        """
        Parameters
        ----------
        hub_height : float | int
            Turbine hub height (m), used along with rotor diameter to
            compute blade tip height which is used to determine setback
            distance.
        rotor_diameter : float | int
            Turbine rotor diameter (m), used along with hub height to
            compute blade tip height which is used to determine setback
            distance.
        regulations_fpath : str | None, optional
            Path to regulations ``.csv`` or ``.gpkg`` file. At a
            minimum, this file must contain the following columns:

                - ``Feature Type``: Contains labels for the type of
                  setback that each row represents. This should be a
                  `"feature_type"` label that can be found in the
                  :attr:`~reVX.setbacks.setbacks.SETBACK_SPECS`
                  dictionary (e.g. ``"structures"``, ``"roads"``,
                  ``"water"``, etc.), unless you have created your own
                  setback calculator using
                  :func:`~reVX.setbacks.setbacks.setbacks_calculator`,
                  in which case this label can match the `feature_type`
                  input you used for that function call.
                - ``Feature Subtype``: Contains labels for feature
                  subtypes. The feature subtypes are only used for
                  down-selecting the local regulations that should be
                  applied for a particular feature, so often you can
                  leave this blank or set it to ``None``. If you do
                  specify this value, it should be a
                  `"feature_subtypes_to_exclude"` label that can be
                  found in the
                  :attr:`~reVX.setbacks.setbacks.SETBACK_SPECS`
                  dictionary,  unless you have created your own setback
                  calculator using
                  :func:`~reVX.setbacks.setbacks.setbacks_calculator`,
                  in which case this label can match the
                  `feature_subtypes_to_exclude` input you used for that
                  function call.
                - ``Value Type``: Specifies wether the value is a
                  multiplier or static height. See below for more info.
                - ``Value``: Numeric value of the setback or multiplier.
                - ``FIPS``: Specifies a unique 5-digit code for each
                  county (this can be an integer - no leading zeros
                  required). This is used to match the county
                  regulations to the county's spatial extent.

            Valid options for the ``Value Type`` are (case-insensitive;
            dashes, underscores, and spaces are interchangeable):

                - "Max-tip Height Multiplier"
                - "Rotor-Diameter Multiplier"
                - "Hub-height Multiplier"
                - "Meters"

            If this input is ``None``, a generic setback of
            ``max_tip_height * multiplier`` is used.
            By default ``None``.
        multiplier : int | float | str | None, optional
            A setback multiplier to use if regulations are not supplied.
            This multiplier will be applied to the ``base_setback_dist``
            to calculate the setback. If supplied along with
            ``regulations_fpath``, this input will be used to apply a
            setback to all counties not listed in the regulations file.
            If this input is a string, it must be a key in
            :attr:`MULTIPLIERS`. By default `None`.
        """
        self._hub_height = hub_height
        self._rotor_diameter = rotor_diameter
        max_tip_height = hub_height + rotor_diameter / 2
        super().__init__(base_setback_dist=max_tip_height,
                         regulations_fpath=regulations_fpath,
                         multiplier=multiplier)

    def _preflight_check(self, regulations_fpath):
        """ Run preflight checks on WindSetbackRegulations inputs.

        In addition to the checks performed in `Regulations`, the
        `multiplier` is converted to a float values if a string is
        input.

        Parameters
        ----------
        regulations_fpath : str | None
            Path to wind regulations .csv or .gpkg file, if None create
            global setbacks.
        """
        super()._preflight_check(regulations_fpath)
        if isinstance(self._multi, str):
            self._multi = self.MULTIPLIERS.get(self._multi)
            logger.debug('Computing setbacks using generic Max-tip Height '
                         'Multiplier of {}'.format(self._multi))

    @property
    def hub_height(self):
        """
        Turbine hub height in meters

        Returns
        -------
        float
        """
        return self._hub_height

    @property
    def rotor_diameter(self):
        """
        Turbine rotor diameter in meters

        Returns
        -------
        float
        """
        return self._rotor_diameter

    def _county_regulation_value(self, county_regulations):
        """Retrieve county regulation setback. """
        setback_type = county_regulations["Value Type"]
        setback = float(county_regulations["Value"])
        if setback_type == "max tip height multiplier":
            setback *= self.base_setback_dist
        elif setback_type == "rotor diameter multiplier":
            setback *= self.rotor_diameter
        elif setback_type == "hub height multiplier":
            setback *= self.hub_height
        elif setback_type != "meters":
            msg = ('Cannot create setback for {}, expecting '
                   '"Max-tip Height Multiplier", '
                   '"Rotor-Diameter Multiplier", '
                   '"Hub-height Multiplier", or '
                   '"Meters", but got {!r}'
                   .format(county_regulations["County"], setback_type))
            logger.warning(msg)
            warn(msg)
            return
        return setback


def validate_setback_regulations_input(base_setback_dist=None, hub_height=None,
                                       rotor_diameter=None):
    """Validate the setback regulations initialization input.

    Specifically, this function raises an error unless exactly one of
    the following combinations of inputs are provided:

        - base_setback_dist
        - hub_height and rotor_diameter


    Parameters
    ----------
    base_setback_dist : float | int
        Base setback distance (m). This value will be used to calculate
        the setback distance when a multiplier is provided either via
        the ``regulations_fpath`` csv or the ``multiplier`` input. In
        these cases, the setbacks will be calculated using
        ``base_setback_dist * multiplier``. By default, `None`.
    hub_height : float | int
        Turbine hub height (m), used along with rotor diameter to
        compute blade tip-height which is used as the base setback
        distance.  By default, ``None``.
    rotor_diameter : float | int
        Turbine rotor diameter (m), used along with hub height to
        compute blade tip-height which is used as the base setback
        distance. By default, ``None``.

    Raises
    ------
    RuntimeError
        If not enough info is provided (all inputs are ``None``), or too
        much info is given (all inputs are not ``None``).
    """
    no_base_setback = base_setback_dist is None
    invalid_turbine_specs = rotor_diameter is None or hub_height is None

    not_enough_info = no_base_setback and invalid_turbine_specs
    too_much_info = not no_base_setback and not invalid_turbine_specs
    if not_enough_info or too_much_info:
        raise RuntimeError("Must provide either `base_setback_dist` or both "
                           "`rotor_diameter` and `hub_height` (but not all "
                           "three).")


def select_setback_regulations(base_setback_dist=None, hub_height=None,
                               rotor_diameter=None, regulations_fpath=None,
                               multiplier=None):
    """Select appropriate setback regulations based on input.

    Parameters
    ----------
    base_setback_dist : float | int
        Base setback distance (m). This value will be used to calculate
        the setback distance when a multiplier is provided either via
        the ``regulations_fpath`` csv or the ``multiplier`` input. In
        these cases, the setbacks will be calculated using
        ``base_setback_dist * multiplier``. By default, `None`.
    hub_height : float | int
        Turbine hub height (m), used along with rotor diameter to
        compute blade tip-height which is used as the base setback
        distance.  By default, ``None``.
    rotor_diameter : float | int
        Turbine rotor diameter (m), used along with hub height to
        compute blade tip-height which is used as the base setback
        distance. By default, ``None``.
    regulations_fpath : str | None, optional
        Path to regulations ``.csv`` or ``.gpkg`` file. At a minimum,
        this file must contain the following columns: ``Feature Type``,
        which contains labels for the type of setback that each row
        represents, ``Value Type``, which specifies wether the value is
        a multiplier or static height, ``Value``, which specifies the
        numeric value of the setback or multiplier, and ``FIPS``, which
        specifies a unique 5-digit code for each county (this can be an
        integer - no leading zeros required). Valid options for the
        ``Value Type`` are (case-insensitive; dashes, underscores,
        and spaces are interchangeable):

            - "Structure Height Multiplier"
            - "Meters"

        If this input is ``None``, a generic setback of
        ``base_setback_dist * multiplier`` is used. By default ``None``.
    multiplier : int | float | str | None, optional
        A setback multiplier to use if regulations are not supplied.
        This multiplier will be applied to the ``base_setback_dist``
        to calculate the setback. If supplied along with
        ``regulations_fpath``, this input will be used to apply a
        setback to all counties not listed in the regulations file.
        By default ``None``.

    Returns
    -------
    Regulations
        A regulations object that can be used to calculate the requested
        setback distance.
    """

    validate_setback_regulations_input(base_setback_dist=base_setback_dist,
                                       hub_height=hub_height,
                                       rotor_diameter=rotor_diameter)

    if base_setback_dist is None:
        return WindSetbackRegulations(hub_height=hub_height,
                                      rotor_diameter=rotor_diameter,
                                      regulations_fpath=regulations_fpath,
                                      multiplier=multiplier)
    else:
        return SetbackRegulations(base_setback_dist=base_setback_dist,
                                  regulations_fpath=regulations_fpath,
                                  multiplier=multiplier)



# -*- coding: utf-8 -*-
"""
Class to handle geotiff input files.

Created on Thu Jun 20 09:43:34 2019

@author: gbuster
"""
import os
import logging
import rasterio
import numpy as np
import pandas as pd
from affine import Affine
from pyproj import Transformer

from rex.utilities.parse_keys import parse_keys
from reVX.utilities.exceptions import GeoTiffKeyError


logger = logging.getLogger(__name__)


class Geotiff:
    """GeoTIFF handler object."""

    def __init__(self, fpath, chunks=(128, 128)):
        """
        Parameters
        ----------
        fpath : str
            Path to .tiff file.
        chunks : tuple
            GeoTIFF chunk (tile) shape/size.
        """
        self._fpath = fpath
        self._iarr = None
        self._src = rasterio.open(self._fpath, chunks=chunks)
        self._profile = dict(self._src.profile)
        self._profile["transform"] = self._profile["transform"][:6]
        self._profile["crs"] = self._profile["crs"].to_proj4()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()
        if type is not None:
            raise

    def __len__(self):
        """Total number of pixels in the GeoTiff."""
        return self.n_rows * self.n_cols

    def __getitem__(self, keys):
        """Retrieve data from the GeoTIFF object.

        Example, get meta data and layer-0 data for rows 0 through 128 and
        columns 128 through 256.

            meta = geotiff['meta', 0:128, 128:256]
            data = geotiff[0, 0:128, 128:256]

        Parameters
        ----------
        keys : tuple
            Slicing args similar to a numpy array slice. See examples above.
        """
        ds, ds_slice = parse_keys(keys)
        out = None
        if isinstance(ds, str):
            if ds == 'meta':
                out = self._get_meta(*ds_slice)
            elif ds.casefold() == 'lat_lon':
                out = self._get_lat_lon(*ds_slice)
            elif ds.lower().startswith('lat'):
                out = self._get_lat_lon(*ds_slice)[0]
            elif ds.lower().startswith('lon'):
                out = self._get_lat_lon(*ds_slice)[1]

        if out is None:
            out = self._get_data(ds, *ds_slice)

        return out

    @property
    def profile(self):
        """
        GeoTiff geospatial profile

        Returns
        -------
        _profile : dict
            Dictionary of geo-spatial attributes needed to create GeoTiff
        """
        return self._profile

    @property
    def dtype(self):
        """
        GeoTiff array dtype

        Returns
        -------
        dtype : str
            Dtype of data in GeoTiff
        """
        return self.profile["dtype"]

    @property
    def iarr(self):
        """Get an array of 1D index values for the flattened geotiff extent.

        Returns
        -------
        iarr : np.ndarray
            Uint array with same shape as geotiff extent, representing the 1D
            index values if the geotiff extent was flattened
            (with default flatten order 'C')
        """
        if self._iarr is None:
            self._iarr = np.arange(len(self), dtype=np.uint32)
            self._iarr = self._iarr.reshape(self.shape)
        return self._iarr

    @property
    def tiff_shape(self):
        """
        Tiff array shape (bands, y, x)

        Returns
        -------
        shape : tuple
            (bands, y, x)
        """
        return (self.bands, *self.shape)

    @property
    def shape(self):
        """Get the Geotiff shape tuple (n_rows, n_cols).

        Returns
        -------
        shape : tuple
            2-entry tuple representing the full GeoTiff shape.
        """
        return self._src.shape

    @property
    def n_rows(self):
        """Get the number of Geotiff rows.

        Returns
        -------
        n_rows : int
            Number of row entries in the full geotiff.
        """
        return self.shape[0]

    @property
    def n_cols(self):
        """Get the number of Geotiff columns.

        Returns
        -------
        n_cols : int
            Number of column entries in the full geotiff.
        """
        return self.shape[1]

    @property
    def bands(self):
        """
        Get number of GeoTiff bands

        Returns
        -------
        bands : int
        """
        return self._src.count

    @property
    def lat_lon(self):
        """
        Get latitude and longitude coordinate arrays

        Returns
        -------
        tuple
        """
        return self._get_lat_lon(slice(None), slice(None))

    @property
    def latitude(self):
        """
        Get latitude coordinates array

        Returns
        -------
        ndarray
        """
        return self['lat']

    @property
    def longitude(self):
        """
        Get longitude coordinates array

        Returns
        -------
        ndarray
        """
        return self['lon']

    @property
    def meta(self):
        """
        Lat lon to y, x coordinate mapping

        Returns
        -------
        pd.DataFrame
        """
        return self['meta']

    @property
    def values(self):
        """
        Full DataArray in [bands, y, x] dimensions

        Returns
        -------
        ndarray
        """
        return self._src.read()

    @staticmethod
    def _unpack_slices(*yx_slice):
        """Get the flattened geotiff layer data.

        Parameters
        ----------
        *yx_slice : tuple
            Slicing args for data

        Returns
        -------
        y_slice : slice
            Row slice.
        x_slice : slice
            Col slice.
        """
        if len(yx_slice) == 1:
            y_slice = yx_slice[0]
            x_slice = slice(None)
        elif len(yx_slice) == 2:
            y_slice = yx_slice[0]
            x_slice = yx_slice[1]
        else:
            raise GeoTiffKeyError('Cannot do 3D slicing on GeoTiff meta.')

        return y_slice, x_slice

    @staticmethod
    def _get_meta_inds(x_slice, y_slice):
        """Get the row and column indices associated with lat/lon slices.

        Parameters
        ----------
        x_slice : slice
            Column slice corresponding to the extracted lon values.
        y_slice : slice
            Row slice corresponding to the extracted lat values.

        Returns
        -------
        row_ind : np.ndarray
            1D array of the row indices corresponding to the lat/lon arrays
            once mesh-gridded and flattened
        col_ind : np.ndarray
            1D array of the col indices corresponding to the lat/lon arrays
            once mesh-gridded and flattened
        """
        if y_slice.start is None:
            y_slice = slice(0, y_slice.stop)
        if x_slice.start is None:
            x_slice = slice(0, x_slice.stop)

        x_len = x_slice.stop - x_slice.start
        y_len = y_slice.stop - y_slice.start

        col_ind = np.arange(x_slice.start, x_slice.start + x_len)
        row_ind = np.arange(y_slice.start, y_slice.start + y_len)
        col_ind = col_ind.astype(np.uint32)
        row_ind = row_ind.astype(np.uint32)
        col_ind, row_ind = np.meshgrid(col_ind, row_ind)
        col_ind = col_ind.flatten()
        row_ind = row_ind.flatten()

        return row_ind, col_ind

    def _get_meta(self, *ds_slice):
        """Get the geotiff meta dataframe in standard WGS84 projection.

        Parameters
        ----------
        *ds_slice : tuple
            Slicing args for meta data.

        Returns
        -------
        meta : pd.DataFrame
            Flattened meta data with same format as reV resource meta data.
        """
        y_slice, x_slice = self._unpack_slices(*ds_slice)
        row_ind, col_ind = self._get_meta_inds(x_slice, y_slice)

        lat, lon = self._get_lat_lon(*ds_slice)
        lon = lon.flatten()
        lat = lat.flatten()

        meta = pd.DataFrame({'latitude': lat.astype(np.float32),
                             'longitude': lon.astype(np.float32),
                             'row_ind': row_ind, 'col_ind': col_ind})
        return meta

    # pylint: disable=all
    def _get_lat_lon(self, *ds_slice):
        """
        Get the geotiff latitude and longitude coordinates

        Parameters
        ----------
        *ds_slice : tuple
            Slicing args for latitude and longitude arrays

        Returns
        -------
        lat : ndarray
            Projected latitude coordinates
        lon : ndarray
            Projected longitude coordinates
        """
        row_slice, col_slice = self._unpack_slices(*ds_slice)

        if all(np.iinfo("uint16").max > dim for dim in self.shape):
            dtype = "uint16"
        else:
            dtype = "uint32"

        rows = np.arange(self.n_rows, dtype=dtype)
        cols = np.arange(self.n_cols, dtype=dtype)

        if any(isinstance(s, slice) for s in [row_slice, col_slice]):
            cols, rows = np.meshgrid(cols[col_slice], rows[row_slice])
        else:
            rows = rows[row_slice]
            cols = cols[col_slice]

        pixel_center_translation = Affine.translation(0.5, 0.5)
        adjusted_transform = self._src.transform * pixel_center_translation
        cols, rows = adjusted_transform * [cols, rows]

        transformer = Transformer.from_crs(self._src.profile["crs"],
                                           'epsg:4326', always_xy=True)

        cols, rows = transformer.transform(cols, rows)
        return rows.astype(np.float32), cols.astype(np.float32)

    def _get_data(self, ds, *ds_slice):
        """Get the flattened geotiff layer data.

        Parameters
        ----------
        ds : int
            Layer to get data from
        *ds_slice : tuple
            Slicing args for data

        Returns
        -------
        data : np.ndarray
            1D array of flattened data corresponding to meta data.
        """
        y_slice, x_slice = self._unpack_slices(*ds_slice)

        if x_slice.stop is None:
            x_slice = slice(x_slice.start, self.shape[1], x_slice.step)
        if y_slice.stop is None:
            y_slice = slice(y_slice.start, self.shape[0], y_slice.step)

        window = rasterio.windows.Window.from_slices(y_slice, x_slice)
        data = self._src.read(ds + 1, window=window).flatten()

        return data

    def close(self):
        """Close the rasterio source object"""
        self._src.close()

    @staticmethod
    def write(out_fp, profile, values, dtype=None):
        """Write values to GeoTIFF file with given profile.

        Parameters
        ----------
        out_fp : str
            Path to GeoTIFF output file to save data to.
        profile : dict
            GeoTIFF profile (attributes).
        values : ndarray
            GeoTIFF data to save.
        dtype : str, optional
            Type of data being stored. If ``None``, the data dtype is
            inferred from the `values` input itself.
        """
        out_dir = os.path.dirname(out_fp)
        if out_dir and not os.path.exists(out_dir):
            logger.debug("Creating %s", out_dir)
            os.makedirs(out_dir)

        if values.ndim < 3:
            values = np.expand_dims(values, 0)

        dtype = dtype or values.dtype.name
        profile['dtype'] = dtype

        if "nodata" not in profile:
            if np.issubdtype(dtype, np.integer):
                dtype_max = np.iinfo(dtype).max
            else:
                dtype_max = np.finfo(dtype).max
            profile['nodata'] = dtype_max

        with rasterio.open(out_fp, 'w', **profile) as f:
            f.write(values)

        logger.debug('%s created', out_fp)



# -*- coding: utf-8 -*-
"""
GeoPackage Handlers
"""
from functools import cached_property

import pyproj
import sqlite3


class GPKGMeta:
    """Extract meta info about underlying GeoPackage SQL tables."""

    def __init__(self, filename):
        """

        Parameters
        ----------
        filename : path-like
            Path to input GeoPackage.
        """
        self.filename = filename

    @cached_property
    def primary_table(self):
        """str: Name of table containing user data."""
        with sqlite3.connect(self.filename) as con:
            cursor = con.cursor()
            cursor.execute("SELECT table_name FROM gpkg_contents")
            return cursor.fetchall()[0][0]

    @cached_property
    def srs_id(self):
        """int: Spatial reference system ID (4326, 3857, etc.)"""
        with sqlite3.connect(self.filename) as con:
            cursor = con.cursor()
            cursor.execute("SELECT srs_id FROM gpkg_contents")
            return cursor.fetchall()[0][0]

    @cached_property
    def crs(self):
        """pyproj.crs.CRS: Coordinate Reference System for GeoPackage."""
        with sqlite3.connect(self.filename) as con:
            cursor = con.cursor()
            cursor.execute("SELECT definition FROM gpkg_spatial_ref_sys "
                           "WHERE srs_id={}".format(self.srs_id))
            proj = cursor.fetchall()[0][0]
            return pyproj.Proj(proj).crs

    @cached_property
    def bbox(self):
        """tuple: MIN_X, MIN_Y, MAX_X, MAX_Y values for GeoPackage. """
        with sqlite3.connect(self.filename) as con:
            cursor = con.cursor()
            cursor.execute("SELECT min_x, min_y, max_x, max_y "
                           "FROM gpkg_contents;")
            return cursor.fetchall()[0]

    @cached_property
    def primary_key_column(self):
        """str: Name of the primary key column in the user data table. """
        with sqlite3.connect(self.filename) as con:
            cursor = con.cursor()
            cursor.execute("PRAGMA table_info({})"
                           .format(self.primary_table))
            pragma = cursor.fetchall()
            pk_name = [info[1] for info in pragma if info[-1]]
            assert len(pk_name) == 1, "Found multiple Primary Key columns"
            return pk_name[0]

    @cached_property
    def geom_table_suffix(self):
        """str: Name of the geometry table suffix."""
        with sqlite3.connect(self.filename) as con:
            cursor = con.cursor()
            cursor.execute("SELECT table_name, column_name FROM "
                           "gpkg_geometry_columns;")
            return "_".join(cursor.fetchall()[0])

    @cached_property
    def num_feats(self):
        """int: Number of feature rows."""
        with sqlite3.connect(self.filename) as con:
            cursor = con.cursor()
            cursor.execute("SELECT COUNT(distinct id) FROM "
                           "rtree_{table_suffix};"
                           .format(table_suffix=self.geom_table_suffix))
            return cursor.fetchall()[0][0]

    @cached_property
    def feat_ids(self):
        """tuple: All the feature ID's in the GeoPackage. """
        with sqlite3.connect(self.filename) as con:
            cursor = con.cursor()
            cursor.execute("PRAGMA temp_store=2")
            cursor.execute("SELECT distinct id FROM rtree_{table_suffix} "
                           "ORDER BY miny, minx"
                           .format(table_suffix=self.geom_table_suffix))
            return tuple(id_[0] for id_ in cursor.fetchall())

    def feat_ids_for_bbox(self, bbox):
        """Find the ID's of the features within the input bounding box.

        Parameters
        ----------
        bbox : list | tuple | iterable
            Container with four int/float values describing the bounding
            box. The values should be in the following order:
            ``MIN_X, MIN_Y, MAX_X, MAX_Y``

        Returns
        -------
        ids : tuple
            Tuple of all the feature IDs within the given input bounding
            box.
        """
        minx, miny, maxx, maxy = bbox
        with sqlite3.connect(self.filename) as con:
            cursor = con.cursor()
            cursor.execute("SELECT distinct id FROM rtree_{table_suffix} "
                           "WHERE minx<={maxx} and maxx>={minx} "
                           "and miny<={maxy} and maxy>={miny} "
                           "ORDER BY miny, minx"
                           .format(table_suffix=self.geom_table_suffix,
                                   maxx=maxx, minx=minx, maxy=maxy, miny=miny))
            ids = tuple(id_[0] for id_ in cursor.fetchall())
        return ids



# -*- coding: utf-8 -*-
"""
Compute setbacks exclusions
"""
from warnings import warn
import logging

from reVX.utilities.regulations import AbstractBaseRegulations


logger = logging.getLogger(__name__)


class FlickerRegulations(AbstractBaseRegulations):
    """Shadow flicker regulation values. """

    def __init__(self, hub_height, rotor_diameter, flicker_threshold=None,
                 regulations_fpath=None):
        """
        Parameters
        ----------
        hub_height : float | int
            Turbine hub height (m).
        rotor_diameter : float | int
            Turbine rotor diameter (m).
        flicker_threshold : float | int, optional
            Maximum number of allowable flicker hours per year to use
            for generic flicker regulations. If `None`, then only local
            (county) flicker regulations are applied.
            By default, `None`.
        regulations_fpath : str, optional
            Path to regulations .csv or .gpkg file. At a minimum, this
            file must contain the following columns: `Feature Type`
            which labels the type of regulation that each row
            represents (flicker regulations must be called "Shadow
            Flicker"), `Value Type`, which specifies the type of the
            value (flicker value types must be "Hrs/Year"), `Value`,
            which specifies the numeric value of the flicker threshold
            (in hours), and `FIPS`, which specifies a unique 5-digit
            code for each county (this can be an integer - no leading
            zeros required). If this input is `None`, generic flicker
            regulations defined by `flicker_threshold` are applied.
            By default `None`.
        """
        self._hub_height = hub_height
        self._rotor_diameter = rotor_diameter
        super().__init__(generic_regulation_value=flicker_threshold,
                         regulations_fpath=regulations_fpath)

    @property
    def hub_height(self):
        """float | int: Turbine hub height in meters. """
        return self._hub_height

    @property
    def rotor_diameter(self):
        """float | int: Turbine rotor diameter in meters. """
        return self._rotor_diameter

    def _county_regulation_value(self, county_regulations):
        """Retrieve county regulation value. """
        regulation_type = county_regulations["Value Type"]
        regulation = float(county_regulations["Value"])
        if regulation_type != "hrs/year":
            msg = ('Cannot create flicker regulations for {}, expecting '
                   '"Hrs/Year", but got {!r}'
                   .format(county_regulations["County"], regulation_type))
            logger.warning(msg)
            warn(msg)
            return
        return regulation



# -*- coding: utf-8 -*-
"""
Helper functions for setback exclusion computation
"""

import geopandas as gpd


def positive_buffer(features, regulation_value):
    """Buffer features using a given regulation value.

    This function applies a simple positive buffer to every polygon in
    the input features.

    Parameters
    ----------
    features : geopandas.GeoDataFrame
        Features to apply buffer to.
    regulation_value : int | float
        Regulations value used to buffer the features. This value should
        be in the same units as the ``features`` input GeoDataFrame.

    See Also
    --------
    geopandas.GeoSeries.buffer : Function used ot buffer each feature.

    Returns
    -------
    list
        List of buffered buffered feature shapes.
    """
    return list(features.buffer(regulation_value))


def parcel_buffer(features, regulation_value):
    """Buffer features imitating a parcel setback.

    Parcel (property-line) setbacks typically prohibit any type of build
    within "x" meters of a property line. Therefore, this type of buffer
    first applies a negative buffer on each input feature and then takes
    the difference between the original and the negatively-buffered
    feature. The resulting shape is the "inside" of the original input
    feature, where each edge is no closer than ``regulation_value`` to
    the original feature boundary.

    Parameters
    ----------
    features : geopandas.GeoDataFrame
        Features to apply buffer to.
    regulation_value : int | float
        Regulations value used to (negatively) buffer the features.
        This value should be in the same units as the ``features`` input
        GeoDataFrame.

    See Also
    --------
    geopandas.GeoSeries.buffer : Function used ot buffer each feature.

    Returns
    -------
    list
        List of buffered buffered feature shapes.
    """
    negative_buffer = features.buffer(-1 * regulation_value)
    return list(features.buffer(0).difference(negative_buffer))


def features_with_centroid_in_county(features, county):
    """Filter features to those with centroids within the given county.

    Parameters
    ----------
    features : geopandas.GeoDataFrame
        Features to setback from.
    county : geopandas.GeoDataFrame
        Regulations for a single county.

    Returns
    -------
    features : geopandas.GeoDataFrame
        Features that have centroid in county.
    """
    mask = features.centroid.within(county['geometry'].values[0])
    return features.loc[mask]


def features_clipped_to_county(features, county):
    """Clip features to the given county geometry.

    Parameters
    ----------
    features : geopandas.GeoDataFrame
        Features to setback from.
    county : geopandas.GeoDataFrame
        Regulations for a single county.

    Returns
    -------
    features : geopandas.GeoDataFrame
        Features clipped to county geometry.
    """
    tmp = gpd.clip(features, county)
    return tmp[~tmp.is_empty]



# -*- coding: utf-8 -*-
"""
Created on Tues Feb 2023

@author: bbenton
"""
import logging
import pandas as pd
import numpy as np
import os
import shutil
from warnings import warn
import re

from reVX.plexos.rev_reeds_plexos import PlexosAggregation
from rex import Outputs


logger = logging.getLogger(__name__)


class PrasAggregation(PlexosAggregation):
    """
    Framework to aggregate reV gen profiles to PRAS zone power profiles and
    overwrite exisiting pras file with new generation profiles. This class
    takes as input rev supply curve table, and reeds buildout table (specifying
    which rev sc points were built and at what capacity). The class will build
    power profiles for each supply curve point and then aggregate the sc point
    profiles to the nearest neighbor pras zone. The corresponding zone power
    profiles in the pras output file will be overwritten with these new
    aggregated generation profiles.

    Examples
    --------

    >>> from reVX.pras.rev_reeds_pras import PrasAggregation
    >>>
    >>> rev_supply_curve_file = ...
    >>> reeds_build_file = ...
    >>> capacity_factor_file = ...
    >>> pras_file = ...
    >>>
    >>> PrasAggregation.run(rev_supply_curve_file, reeds_build_file,
                            capacity_factor_file, pras_file, build_year=2047,
                            tech_type='upv')

    This will replace all the pras generation profiles with upv tech in the
    file pras_file.pras.h5 with aggregated generation profiles from cf_path.h5

    >>> reeds_build_files = [...]
    >>> capacity_factor_files = [...]
    >>> tech_types = ['upv', 'dupv', 'wind-ons', 'wind-ofs']
    >>>
    >>> PrasAggregation.run(rev_supply_curve_file, reeds_build_files,
                            capacity_factor_files, pras_file, build_year=2047,
                            tech_type=tech_types)

    This will run the same routine but for each tech type in tech_types. The
    files in reeds_build_files and capacity_factor_files have to correspond to
    the tech type in the tech_types list.
    """
    def __init__(self, rev_sc, reeds_build, cf_fpath, pras_file,
                 forecast_fpath=None, build_year=2050, res_class=None,
                 tech_type=None, timezone='US/Central', dset_tag=None,
                 max_workers=None):
        """
        Parameters
        ----------
        rev_sc : str | pd.DataFrame
            reV supply curve results table including SC gid, latitude,
            longitude, res_gids, gid_counts. Or file path to reV supply
            curve table.
        reeds_build : str | pd.DataFrame
            ReEDS buildout with rows for built capacity (MW) at each reV SC
            point. This should have columns: year, region, class,
            built_capacity, and sc_gid (corresponding to the reV supply curve
            point gid). Some cleaning of the column names will be performed for
            legacy tables but these are the column headers that are desired.
        cf_fpath : str
            File path to .h5 capacity factor file (reV gen output) to get
            profiles from.
        pras_file : str
            File path to existing hdf5 PRAS file (.pras or .h5 extension) which
            will be amended to include generation data from given resource
            data. This file must have a 'generators' group with '_core' and
            'capacity' datasets in this group.
        forecast_fpath : str | None
            Forecasted capacity factor .h5 file path (reV results). If not
            None, the supply curve res_gids are mapped to sites in the
            cf_fpath, then the coordinates from cf_fpath are mapped to the
            nearest neighbor sites in the forecast_fpath, where the final
            generation profiles are retrieved from.
        build_year : int, optional
            REEDS year of interest, by default 2050
        res_class : int | None
            Optional resource class to use to filter supply curve points.
            For example, if res_class = 3 then only supply curve points with
            class 3 will be kept in the sc_build table. The corresponds to the
            'class' column in the reeds_build file.
        tech_type : str | None
            Pras files will have the tech type in the
            ['generation/_core']['name'] entry, which is used to select the
            correct output index
        timezone : str
            Timezone for output generation profiles. This is a string that will
            be passed to pytz.timezone() e.g. US/Pacific, US/Mountain,
            US/Central, US/Eastern, or UTC. For a list of all available
            timezones, see pytz.all_timezones
        dset_tag : str
            Dataset tag to append to dataset names in cf profile file. e.g. If
            the cf profile file is a multi year file using dset_tag="-2008"
            will enable us to select the corresponding datasets
            (cf_mean-2008, cf_profile-2008, etc)
        max_workers : int | None
            Max workers for parallel profile aggregation. None uses all
            available workers. 1 will run in serial.
        """
        self._dset_tag = dset_tag if dset_tag is not None else ""
        self._tech_type = tech_type
        self._pras_file = pras_file
        self._output_file = pras_file
        self._pras_meta = None
        self._res_class = res_class
        self._complete_pras_meta = None
        self._build_year = build_year
        self._sc_build_indices = None
        self._pras_indices = None
        self._qa_file = None
        self._prof_file = None
        self._pras_build_zones = self._init_pras_build_zones(reeds_build)
        super().__init__(self._pras_build_zones, rev_sc,
                         reeds_build, cf_fpath,
                         forecast_fpath=forecast_fpath,
                         build_year=build_year,
                         res_class=res_class,
                         timezone=timezone,
                         dset_tag=self._dset_tag,
                         max_workers=max_workers)
        self._init_output_files(pras_file)
        logger.info('Running aggregation for tech_type={}, res_class={}, '
                    'rev_sc={}, reeds_build={}, cf_fpath={}, dset_tag={}, '
                    'pras_file={}.'
                    .format(self._tech_type, self._res_class, rev_sc,
                            reeds_build, cf_fpath, dset_tag, pras_file))
        logger.info('Found {} pras_zones and {} sc_build_zones'
                    .format(len(self.pras_zones), len(self.sc_build_zones)))
        self.missing_zones = self.get_missing_zones()
        self.found_zones = list(set(self.pras_zones) - set(self.missing_zones))

    def _init_pras_build_zones(self, reeds_build):
        """
        Build pras_build_zones table from reeds_build. This returns a table
        with all available zones in the reeds build table and assigns a
        plexos_id to each zone so that this can be used as a plexos node table
        for calling super().__init__()

        Parameters
        ----------
        reeds_build : str | pd.DataFrame
            ReEDS buildout with rows for built capacity (MW) at each reV SC
            point. This should have columns: reeds_year, built_capacity, and
            sc_gid (corresponding to the reV supply curve point gid). Some
            cleaning of the column names will be performed for legacy tables
            but these are the column headers that are desired.

        Returns
        -------
        pras_build_zones : pd.DataFrame
            pras zones DataFrame with all available zones in reeds build table
            and corresponding plexos_id
        """
        pras_build_zones = reeds_build
        if isinstance(pras_build_zones, str):
            pras_build_zones = pd.read_csv(pras_build_zones)
        class_mask = pras_build_zones['class'] == self._res_class
        pras_build_zones = pras_build_zones[class_mask]
        year_mask = pras_build_zones['year'] == self._build_year
        pras_build_zones = pras_build_zones[year_mask]
        pras_build_zones = pras_build_zones[['latitude', 'longitude',
                                             'region', 'built_capacity']]
        pras_build_zones_grp = pras_build_zones.groupby(by='region')
        pras_build_zones = pras_build_zones_grp.mean()
        cap = pras_build_zones_grp.sum()['built_capacity']
        pras_build_zones['built_capacity'] = cap
        pras_build_zones = pras_build_zones.reset_index()
        pras_build_zones['plexos_id'] = np.arange(len(pras_build_zones))
        return pras_build_zones

    def _init_output_files(self, pras_file):
        """
        Initialize the output file to be written with new generation profiles
        and define the qa file name.

        Parameters
        ----------
        pras_file : str
            Path to original pras file containing generation profiles to be
            updated. The file will be copied to a new path (given by
            _output_file) and then updated with the new profiles
        """
        if not self._pras_file.endswith('.h5'):
            self._output_file += '.h5'
        self._output_file = pras_file.replace('.h5', '_updated.h5')
        msg = ("pras_file={} does not exist. This file must come from a "
               "previous ReEDS2PRAS run").format(self._pras_file)
        assert os.path.exists(self._pras_file), msg

        if not os.path.exists(self._output_file):
            shutil.copy(self._pras_file, self._output_file)

        self._qa_file = self._output_file.replace(
            '.h5', '_qa_{}_{}.csv'.format(self._tech_type, self._res_class))

        self._prof_file = self._output_file.replace(
            '.h5', '_{}_{}.csv'.format(self._tech_type, self._res_class))

    @classmethod
    def get_pras_meta(cls, pras_file):
        """
        Get pras 'meta' which describes the tech type and class at each zone
        For upv and dupv regions are labeled p1, p2, etc. For wind and csp
        regions are labeled s1, s2, etc.

        Parameters
        ----------
        pras_file : str
            Path to the pras file being overwritten with new generation
            profiles

        Returns
        -------
        pd.DataFrame
        """
        def zone_filter(x):
            out = re.search(r'p\d+', x)
            if out is None:
                out = re.search(r's\d+', x)
            return out[0]

        with Outputs(pras_file, mode='r') as out:
            pras_meta = pd.DataFrame(out['generators/_core'])
            for col in pras_meta.columns:
                pras_meta[col] = pras_meta[col].apply(
                    lambda x: x.decode('utf-8'))

        pras_meta['region'] = pras_meta['name'].apply(zone_filter).values
        return pras_meta

    @classmethod
    def get_pras_classes(cls, pras_file, tech_type):
        """
        Get list of tech classes present in the pras file for the requested
        technology type

        Parameters
        ----------
        pras_file : str
            Path to the pras file being overwritten with new generation
            profiles
        tech_type : str
            Technology type for which profiles are being overwritten

        Returns
        -------
        List
        """
        pras_meta = cls.get_pras_meta(pras_file)
        class_list = [int(x.split('_')[-1]) for x in pras_meta['category']
                      if x.startswith(tech_type)]
        class_list = np.unique(class_list)
        logger.info('Found classes {} for tech_type {}'
                    .format(class_list, tech_type))
        return class_list

    @property
    def zone_index_map(self):
        """
        Get the map between zone index and zone label. i.e. zone_index=1,
        zone_label=p10
        """
        zone_index_map = dict(
            zip(self._pras_build_zones['region'].values,
                self._pras_build_zones['plexos_id'].values))
        return zone_index_map

    @property
    def sc_build_zones(self):
        """
        Get the labels for the zones in the zone_map. For example, the zone_map
        might have an index of 1 for the region but this corresponds to the
        label of p10.
        """
        build_zones = self._pras_build_zones['region'].unique()
        return build_zones

    @property
    def complete_pras_meta(self):
        """
        Get pras 'meta' without filtering for tech_type and class

        Returns
        -------
        pd.DataFrame
        """
        if self._complete_pras_meta is None:
            self._complete_pras_meta = self.get_pras_meta(self._pras_file)
        return self._complete_pras_meta

    @property
    def pras_meta(self):
        """
        Get pras 'meta' which describes the tech type and class at each zone

        Returns
        -------
        pd.DataFrame
        """
        if self._pras_meta is None:
            self._pras_meta = self.enforce_tech_constraint(
                self.complete_pras_meta)
        return self._pras_meta

    @property
    def pras_zones(self):
        """
        Get list of all available zones in pras output file for the requested
        tech and class. For upv and dupv regions are labeled p1, p2, etc. For
        wind and csp regions are labeled s1, s2, etc.

        Returns
        -------
        List
        """
        return self.pras_meta['region'].values

    @property
    def pras_indices(self):
        """
        Get the list of pras file indices for the pras zones available in the
        supply curve zone list.

        Returns
        -------
        np.ndarray
        """
        if self._pras_indices is None:
            zone_mask = np.isin(self.pras_zones, self.sc_build_zones)
            self._pras_indices = list(self.pras_meta.index[zone_mask])
        return self._pras_indices

    @property
    def pras_names(self):
        """
        Get the list of pras entry names for the pras zones available in the
        supply curve zone list.

        Returns
        -------
        np.ndarray
        """
        return self.pras_meta['name'].loc[self.pras_indices]

    @property
    def found_pras_zones(self):
        """
        List of pras zones found in the reeds build out for the requested tech
        and class

        Returns
        -------
        np.ndarray
        """
        return self.pras_meta['region'].loc[self.pras_indices]

    @property
    def found_build_zones(self):
        """
        List of sc build zones found in the pras output file for the requested
        tech and class

        Returns
        -------
        np.ndarray
        """
        return self.sc_build_zones[self.sc_build_indices]

    @property
    def sc_build_indices(self):
        """
        Get the list of profile indices corresponding to the sc_build_zones -
        used to write the new profiles to the pras output.

        Returns
        -------
        List
        """
        if self._sc_build_indices is None:
            zones = self.sc_build_zones.tolist()
            self._sc_build_indices = [zones.index(x) for x
                                      in self.found_pras_zones]
        return self._sc_build_indices

    @property
    def built_capacity(self):
        """
        Get the built capacity for all zones found in the pras output file and
        the reeds build file.

        Returns
        -------
        np.ndarray
        """
        indices = [np.where(self._pras_build_zones['region'] == x)[0][0]
                   for x in self.found_pras_zones]
        cap = self._pras_build_zones['built_capacity'].iloc[indices].values
        return cap

    def enforce_tech_constraint(self, meta):
        """
        Filter the pras generator meta for the requested technology type and
        class

        Parameters
        ----------
        meta : pd.DataFrame
            Full pras 'meta' without tech_type or class filtering

        Returns
        -------
        pd.DataFrame
            pras meta DataFrame filtered by tech_type and class
        """
        tech_key = '{}_'.format(self._tech_type)
        if self._res_class is not None:
            tech_key += '{}'.format(self._res_class)
        category_mask = [x.startswith(tech_key) for x in meta['category']]
        out_meta = meta[category_mask]
        if out_meta.empty:
            msg = ('Found zero pras zones with requested tech_key {}.'
                   .format(tech_key))
            logger.error(msg)
            raise RuntimeError(msg)
        return out_meta

    def get_missing_zones(self):
        """
        Check for zones present in the pras output file which cannot be found
        in the supply curve zone list
        """
        zone_mask = np.isin(self.pras_zones, self.sc_build_zones)
        missing_zones = sorted([x for x in np.unique(self.pras_zones)
                                if x not in self.sc_build_zones])
        msg = ('{} / {} pras zones not found in sc_build_zones, for '
               'tech_key {}_{}. Missing zones: {}.').format(
                   len(missing_zones), len(self.pras_zones), self._tech_type,
                   self._res_class, missing_zones)
        if not all(zone_mask):
            logger.warning(msg)
            warn(msg)
        else:
            logger.info('All {} pras zones found in sc_build_zones'
                        .format(len(self.pras_zones)))

        if len(missing_zones) > 0:
            missing_indices = self.pras_meta['region'].isin(missing_zones)
            missing_indices = self.pras_meta.index[missing_indices].tolist()
            with Outputs(self._pras_file, mode='r') as out:
                missing_cap = out['generators/capacity', :, missing_indices]
                logger.info('The mean/min/max capacity for the missing zones: '
                            '{} / {} / {} MW.'
                            .format(np.mean(missing_cap), np.min(missing_cap),
                                    np.max(missing_cap)))

        return missing_zones

    def qa(self, profiles):
        """
        Simple QA to check original mean/min/max capacity and the mean/min/max
        capacity for the updated generation profiles

        Parameters
        ----------
        profiles : np.ndarray
            Generation profile timeseries in MW at each plant / pras zone.

        Returns
        -------
        new_cap : np.ndarray
            Array containing the updated generation profiles
        """
        df = pd.DataFrame()
        with Outputs(self._pras_file, mode='r') as out:
            old_cap = out['generators/capacity', :, self.pras_indices]
            new_cap = profiles[:, self.sc_build_indices].astype(old_cap.dtype)
            df['region'] = self.found_pras_zones
            df['name'] = self.pras_meta['name'].loc[self.pras_indices]
            df['old_mean'] = np.mean(old_cap, axis=0)
            df['new_mean'] = np.mean(new_cap, axis=0)
            df['old_max'] = np.max(old_cap, axis=0)
            df['new_max'] = np.max(new_cap, axis=0)
            df['built_capacity'] = self.built_capacity
            logger.info('The old and new mean/max capacity (MW) for the {}_{} '
                        'zones:\n{}.'.format(self._tech_type,
                                             self._res_class, df))
            df.to_csv(self._qa_file)
            logger.info('Wrote pras QA to: {}'.format(self._qa_file))

            return new_cap

    def export(self, time_index, profiles):
        """Export generation profiles to h5 and plexos-formatted csv

        Parameters
        ----------
        time_index : pd.datetimeindex
            Time index for the profiles.
        profiles : np.ndarray
            Generation profile timeseries in MW at each plant / pras zone.
        """
        logger.info('Saving profiles for {}_{} to file: {}'
                    .format(self._tech_type, self._res_class,
                            self._output_file))

        profiles = self.tz_convert_profiles(profiles, self._timezone)

        new_cap = self.qa(profiles)
        with Outputs(self._output_file, mode='a') as out:
            out['generators/capacity', :, self.pras_indices] = new_cap

        df_pras = pd.DataFrame(new_cap, columns=self.pras_names,
                               index=time_index.tz_convert(None))
        df_pras.index.name = 'DATETIME'
        df_pras.to_csv(self._prof_file)
        logger.info('Wrote pras formatted profiles to: {}'
                    .format(self._prof_file))

    def _make_node_map(self):
        """Map built rev SC points to pras zones.

        Returns
        -------
        pras_zone_index : np.ndarray
            (n, 1) array of pras zone indices mapped to the SC builds where n
            is the number of SC points built.  Each value in this array gives
            the pras zone index that the sc point is mapped to. So
            self.zone_map[10] yields the pras zone index for
            self.sc_build[10].
        """
        if 'region' in self.sc_build:
            pras_zone_index = self.sc_build['region'].values
            pras_zone_index = [self.zone_index_map[k] for k
                               in pras_zone_index]
            pras_zone_index = np.array(pras_zone_index)

            if len(pras_zone_index.shape) == 1:
                pras_zone_index = pras_zone_index.reshape(
                    (len(pras_zone_index), 1))
        else:
            msg = '"region" column not found in sc_build.'
            logger.error(msg)
            raise RuntimeError(msg)
        return pras_zone_index

    @classmethod
    def run(cls, rev_sc, reeds_build, cf_fpath, pras_file,
            forecast_fpath=None, build_year=2050, res_class=None,
            tech_type=None, timezone='US/Central', dset_tag=None,
            max_workers=None):
        """Run pras aggregation and output for the requested tech types. This
        will aggregate the generation profiles over region specific supply
        curve points found in rev_sc, after filtering rev_sc for the requested
        tech type, and resource class. These aggregated generation profiles
        will then be written to the pras output file in the location
        corresponding to the region, tech type, and class.

        rev_sc : str | pd.DataFrame
            reV supply curve results table including SC gid, latitude,
            longitude, res_gids, gid_counts. Or file path to reV supply
            curve table. If providing a list of tech types this also needs to
            be a list of the corresponding supply curve files.
        reeds_build : str | pd.DataFrame | list
            ReEDS buildout with rows for built capacity (MW) at each reV SC
            point. This should have columns: reeds_year, built_capacity, and
            sc_gid (corresponding to the reV supply curve point gid). Some
            cleaning of the column names will be performed for legacy tables
            but these are the column headers that are desired. This input can
            also include "plexos_node_gid" which will explicitly assign a
            supply curve point buildout to a single pras zone. If included,
            all points must be assigned to pras zones. If providing a list
            of tech types this also needs to be a list of the corresponding
            reeds builout files for each tech type.
        cf_fpath : str | list
            File path to capacity factor file (reV gen output) to
            get profiles from. If providing a list of tech types this also
            needs to be a list of the corresponding capacity factor files.
        pras_file : str
            File path to existing PRAS file which will be amended to include
            generation data from given resource data.
        forecast_fpath : str | None
            Forecasted capacity factor .h5 file path (reV results). If not
            None, the supply curve res_gids are mapped to sites in the
            cf_fpath, then the coordinates from cf_fpath are mapped to the
            nearest neighbor sites in the forecast_fpath, where the final
            generation profiles are retrieved from.
        build_year : int, optional
            REEDS year of interest, by default 2050
        res_class : int | None
            Optional resource class to use to filter supply curve points.
            For example, if res_class = 3 then only supply curve points with
            class 3 will be kept in the sc_build table. If None then the
            aggregation will be run for each class found in the pras output
            file
        tech_type : str | list
            Tech type for which to overwrite the generation profiles in the
            pras output file. Pras files will have the tech type in the
            ['generation/_core']['name'] entry, which is used to select the
            correct output index
        timezone : str
            Timezone for output generation profiles. This is a string that will
            be passed to pytz.timezone() e.g. US/Pacific, US/Mountain,
            US/Central, US/Eastern, or UTC. For a list of all available
            timezones, see pytz.all_timezones
        dset_tag : str
            Dataset tag to append to dataset names in cf profile file. e.g. If
            the cf profile file is a multi year file using dset_tag="-2008"
            will enable us to select the corresponding datasets
            (cf_mean-2008, cf_profile-2008, etc)
        max_workers : int | None
            Max workers for parallel profile aggregation. None uses all
            available workers. 1 will run in serial.
        """

        if not isinstance(reeds_build, list):
            reeds_build = [reeds_build]
        if not isinstance(cf_fpath, list):
            cf_fpath = [cf_fpath]
        if not isinstance(rev_sc, list):
            rev_sc = [rev_sc]
        if not isinstance(tech_type, list):
            tech_type = [tech_type]
        missing_zones = []

        for i, tech in enumerate(tech_type):
            reeds_build_fp = reeds_build[i]
            cf_fp = cf_fpath[i]
            rev_sc_fp = rev_sc[i]
            missing_zones += cls._run_single(tech, rev_sc_fp, reeds_build_fp,
                                             cf_fp, pras_file,
                                             forecast_path=forecast_fpath,
                                             build_year=build_year,
                                             res_class=res_class,
                                             timezone=timezone,
                                             dset_tag=dset_tag,
                                             max_workers=max_workers)

        logger.info('Missing {} zones across all classes: {}'
                    .format(len(set(missing_zones)), set(missing_zones)))

    @classmethod
    def _run_single(cls, tech, rev_sc_fp, reeds_build_fp, cf_fp, pras_file,
                    forecast_fpath=None, build_year=2050, res_class=None,
                    timezone='US/Central', dset_tag=None, max_workers=None):
        """Run pras aggregation and output for a single requested tech type.
        This will aggregate the generation profiles over region specific supply
        curve points found in rev_sc, after filtering rev_sc for the requested
        tech type, and resource class. These aggregated generation profiles
        will then be written to the pras output file in the location
        corresponding to the region, tech type, and class.

        Parameters
        ----------
        tech : str
            Tech type for which to overwrite the generation profiles in the
            pras output file. Pras files will have the tech type in the
            ['generation/_core']['name'] entry, which is used to select the
            correct output index
        rev_sc_fp : str | pd.DataFrame
            reV supply curve results table including SC gid, latitude,
            longitude, res_gids, gid_counts. Or file path to reV supply
            curve table.
        reeds_build_fp : str | pd.DataFrame
            ReEDS buildout with rows for built capacity (MW) at each reV SC
            point. This should have columns: reeds_year, built_capacity, and
            sc_gid (corresponding to the reV supply curve point gid). Some
            cleaning of the column names will be performed for legacy tables
            but these are the column headers that are desired. This input can
            also include "plexos_node_gid" which will explicitly assign a
            supply curve point buildout to a single pras zone. If included,
            all points must be assigned to pras zones.
        cf_fp : str
            File path to capacity factor file (reV gen output) to
            get profiles from.
        pras_file : str
            File path to existing PRAS file which will be amended to include
            generation data from given resource data.
        forecast_fpath : str | None
            Forecasted capacity factor .h5 file path (reV results). If not
            None, the supply curve res_gids are mapped to sites in the
            cf_fpath, then the coordinates from cf_fpath are mapped to the
            nearest neighbor sites in the forecast_fpath, where the final
            generation profiles are retrieved from.
        build_year : int, optional
            REEDS year of interest, by default 2050
        res_class : int | None
            Optional resource class to use to filter supply curve points.
            For example, if res_class = 3 then only supply curve points with
            class 3 will be kept in the sc_build table. If None then the
            aggregation will be run for each class found in the pras output
            file
        timezone : str
            Timezone for output generation profiles. This is a string that will
            be passed to pytz.timezone() e.g. US/Pacific, US/Mountain,
            US/Central, US/Eastern, or UTC. For a list of all available
            timezones, see pytz.all_timezones
        dset_tag : str
            Dataset tag to append to dataset names in cf profile file. e.g. If
            the cf profile file is a multi year file using dset_tag="-2008"
            will enable us to select the corresponding datasets
            (cf_mean-2008, cf_profile-2008, etc)
        max_workers : int | None
            Max workers for parallel profile aggregation. None uses all
            available workers. 1 will run in serial.
        """

        missing_zones = []
        if res_class is None:
            res_classes = cls.get_pras_classes(pras_file, tech)
        else:
            res_classes = [res_class]
        for res_class in res_classes:
            pa = cls(rev_sc_fp, reeds_build_fp, cf_fp, pras_file,
                     forecast_fpath=forecast_fpath,
                     build_year=build_year,
                     tech_type=tech,
                     res_class=res_class,
                     timezone=timezone,
                     dset_tag=dset_tag,
                     max_workers=max_workers)
            profiles = pa.make_profiles()
            pa.export(pa.time_index, profiles)
            missing_zones += pa.missing_zones
        return missing_zones



# -*- coding: utf-8 -*-
"""
Transmission Least Cost Defaults
"""
import os

from reVX.least_cost_xmission.config.xmission_config import (
    XmissionConfig, LandUseClasses, LandUseMultipliers, SlopeMultipliers,
    IsoMultipliers
)
